# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=too-many-lines
"""
This module defines the classes related to band structures or dispersions
in a Brillouin zone, and how to operate on them.
"""
import json
from string import Template

import numpy

from aiida.common.exceptions import ValidationError
from aiida.common.utils import join_labels, prettify_labels

from .kpoints import KpointsData

__all__ = ('BandsData', 'find_bandgap')


def prepare_header_comment(uuid, plot_info, comment_char='#'):
    """Prepare the header."""
    from aiida import get_file_header

    filetext = []
    filetext += get_file_header(comment_char='').splitlines()
    filetext.append('')
    filetext.append(f'Dumped from BandsData UUID={uuid}')
    filetext.append('\tpoints\tbands')
    filetext.append('\t{}\t{}'.format(*plot_info['y'].shape))
    filetext.append('')
    filetext.append('\tlabel\tpoint')
    for label in plot_info['raw_labels']:
        filetext.append(f'\t{label[1]}\t{label[0]:.8f}')

    return '\n'.join(f'{comment_char} {line}' for line in filetext)


def find_bandgap(bandsdata, number_electrons=None, fermi_energy=None):
    """
    Tries to guess whether the bandsdata represent an insulator.
    This method is meant to be used only for electronic bands (not phonons)
    By default, it will try to use the occupations to guess the number of
    electrons and find the Fermi Energy, otherwise, it can be provided
    explicitely.
    Also, there is an implicit assumption that the kpoints grid is
    "sufficiently" dense, so that the bandsdata are not missing the
    intersection between valence and conduction band if present.
    Use this function with care!

    :param number_electrons: (optional, float) number of electrons in the unit cell
    :param fermi_energy: (optional, float) value of the fermi energy.

    :note: By default, the algorithm uses the occupations array
      to guess the number of electrons and the occupied bands. This is to be
      used with care, because the occupations could be smeared so at a
      non-zero temperature, with the unwanted effect that the conduction bands
      might be occupied in an insulator.
      Prefer to pass the number_of_electrons explicitly

    :note: Only one between number_electrons and fermi_energy can be specified at the
      same time.

    :return: (is_insulator, gap), where is_insulator is a boolean, and gap a
             float. The gap is None in case of a metal, zero when the homo is
             equal to the lumo (e.g. in semi-metals).
    """

    # pylint: disable=too-many-return-statements,too-many-branches,too-many-statements,no-else-return

    def nint(num):
        """
        Stable rounding function
        """
        if num > 0:
            return int(num + .5)
        return int(num - .5)

    if fermi_energy and number_electrons:
        raise ValueError('Specify either the number of electrons or the Fermi energy, but not both')

    try:
        stored_bands = bandsdata.get_bands()
    except KeyError:
        raise KeyError('Cannot do much of a band analysis without bands')

    if len(stored_bands.shape) == 3:
        # I write the algorithm for the generic case of having both the spin up and spin down array
        # put all spins on one band per kpoint
        bands = numpy.concatenate(stored_bands, axis=1)
    else:
        bands = stored_bands

    # analysis on occupations:
    if fermi_energy is None:

        num_kpoints = len(bands)

        if number_electrons is None:
            try:
                _, stored_occupations = bandsdata.get_bands(also_occupations=True)
            except KeyError:
                raise KeyError("Cannot determine metallicity if I don't have either fermi energy, or occupations")

            # put the occupations in the same order of bands, also in case of multiple bands
            if len(stored_occupations.shape) == 3:
                # I write the algorithm for the generic case of having both the
                # spin up and spin down array

                # put all spins on one band per kpoint
                occupations = numpy.concatenate(stored_occupations, axis=1)
            else:
                occupations = stored_occupations

            # now sort the bands by energy
            # Note: I am sort of assuming that I have an electronic ground state

            # sort the bands by energy, and reorder the occupations accordingly
            # since after joining the two spins, I might have unsorted stuff
            bands, occupations = [
                numpy.array(y) for y in zip(
                    *[
                        list(zip(*j)) for j in [
                            sorted(zip(i[0].tolist(), i[1].tolist()), key=lambda x: x[0])
                            for i in zip(bands, occupations)
                        ]
                    ]
                )
            ]
            number_electrons = int(round(sum([sum(i) for i in occupations]) / num_kpoints))

            homo_indexes = [numpy.where(numpy.array([nint(_) for _ in x]) > 0)[0][-1] for x in occupations]
            if len(set(homo_indexes)) > 1:  # there must be intersections of valence and conduction bands
                return False, None

            homo = [_[0][_[1]] for _ in zip(bands, homo_indexes)]
            try:
                lumo = [_[0][_[1] + 1] for _ in zip(bands, homo_indexes)]
            except IndexError:
                raise ValueError(
                    'To understand if it is a metal or insulator, '
                    'need more bands than n_band=number_electrons'
                )

        else:
            bands = numpy.sort(bands)
            number_electrons = int(number_electrons)

            # find the zero-temperature occupation per band (1 for spin-polarized
            # calculation, 2 otherwise)
            number_electrons_per_band = 4 - len(stored_bands.shape)  # 1 or 2
            # gather the energies of the homo band, for every kpoint
            homo = [i[number_electrons // number_electrons_per_band - 1] for i in bands]  # take the nth level
            try:
                # gather the energies of the lumo band, for every kpoint
                lumo = [i[number_electrons // number_electrons_per_band] for i in bands]  # take the n+1th level
            except IndexError:
                raise ValueError(
                    'To understand if it is a metal or insulator, '
                    'need more bands than n_band=number_electrons'
                )

        if number_electrons % 2 == 1 and len(stored_bands.shape) == 2:
            # if #electrons is odd and we have a non spin polarized calculation
            # it must be a metal and I don't need further checks
            return False, None

        # if the nth band crosses the (n+1)th, it is an insulator
        gap = min(lumo) - max(homo)
        if gap == 0.:
            return False, 0.

        if gap < 0.:
            return False, None

        return True, gap

    # analysis on the fermi energy
    else:
        # reorganize the bands, rather than per kpoint, per energy level

        # I need the bands sorted by energy
        bands.sort()

        levels = bands.transpose()
        max_mins = [(max(i), min(i)) for i in levels]

        if fermi_energy > bands.max():
            raise ValueError("The Fermi energy is above all band energies, don't know what to do")
        if fermi_energy < bands.min():
            raise ValueError("The Fermi energy is below all band energies, don't know what to do.")

        # one band is crossed by the fermi energy
        if any(i[1] < fermi_energy and fermi_energy < i[0] for i in max_mins):  # pylint: disable=chained-comparison
            return False, None

        # case of semimetals, fermi energy at the crossing of two bands
        # this will only work if the dirac point is computed!
        if (any(i[0] == fermi_energy for i in max_mins) and any(i[1] == fermi_energy for i in max_mins)):
            return False, 0.

        # insulating case, take the max of the band maxima below the fermi energy
        homo = max([i[0] for i in max_mins if i[0] < fermi_energy])
        # take the min of the band minima above the fermi energy
        lumo = min([i[1] for i in max_mins if i[1] > fermi_energy])
        gap = lumo - homo
        if gap <= 0.:
            raise Exception('Something wrong has been implemented. Revise the code!')
        return True, gap


class BandsData(KpointsData):
    """
    Class to handle bands data
    """

    def set_kpointsdata(self, kpointsdata):
        """
        Load the kpoints from a kpoint object.
        :param kpointsdata: an instance of KpointsData class
        """
        if not isinstance(kpointsdata, KpointsData):
            raise ValueError('kpointsdata must be of the KpointsData class')
        try:
            self.cell = kpointsdata.cell
        except AttributeError:
            pass
        try:
            self.pbc = kpointsdata.pbc
        except AttributeError:
            pass
        try:
            the_kpoints = kpointsdata.get_kpoints()
        except AttributeError:
            the_kpoints = None
        try:
            the_weights = kpointsdata.get_kpoints(also_weights=True)[1]
        except AttributeError:
            the_weights = None
        self.set_kpoints(the_kpoints, weights=the_weights)
        try:
            self.labels = kpointsdata.labels
        except (AttributeError, TypeError):
            self.labels = []

    def _validate_bands_occupations(self, bands, occupations=None, labels=None):
        """
        Validate the list of bands and of occupations before storage.
        Kpoints must be set in advance.
        Bands and occupations must be convertible into arrays of
        Nkpoints x Nbands floats or Nspins x Nkpoints x Nbands; Nkpoints must
        correspond to the number of kpoints.
        """
        # pylint: disable=too-many-branches
        try:
            kpoints = self.get_kpoints()
        except AttributeError:
            raise AttributeError('Must first set the kpoints, then the bands')

        the_bands = numpy.array(bands)

        if len(the_bands.shape) not in [2, 3]:
            raise ValueError(
                'Bands must be an array of dimension 2'
                '([N_kpoints, N_bands]) or of dimension 3 '
                ' ([N_arrays, N_kpoints, N_bands]), found instead {}'.format(len(the_bands.shape))
            )

        list_of_arrays_to_be_checked = []

        # check that the shape of everything is consistent with the kpoints
        num_kpoints_from_bands = the_bands.shape[0] if len(the_bands.shape) == 2 else the_bands.shape[1]
        if num_kpoints_from_bands != len(kpoints):
            raise ValueError('There must be energy values for every kpoint')

        if occupations is not None:
            the_occupations = numpy.array(occupations)
            if the_occupations.shape != the_bands.shape:
                raise ValueError(
                    f'Shape of occupations {the_occupations.shape} different from shapeshape of bands {the_bands.shape}'
                )

            if not the_bands.dtype.type == numpy.float64:
                list_of_arrays_to_be_checked.append([the_occupations, 'occupations'])
        else:
            the_occupations = None
            # list_of_arrays_to_be_checked = [ [the_bands,'bands'] ]

        # check that there every element is a float
        if not the_bands.dtype.type == numpy.float64:
            list_of_arrays_to_be_checked.append([the_bands, 'bands'])

        for x, msg in list_of_arrays_to_be_checked:
            try:
                [float(_) for _ in x.flatten() if _ is not None]
            except (TypeError, ValueError):
                raise ValueError(f'The {msg} array can only contain float or None values')

        # check the labels
        if labels is not None:
            if isinstance(labels, str):
                the_labels = [str(labels)]
            elif isinstance(labels, (tuple, list)) and all(isinstance(_, str) for _ in labels):
                the_labels = [str(_) for _ in labels]
            else:
                raise ValidationError(
                    'Band labels have an unrecognized type ({})'
                    'but should be a string or a list of strings'.format(labels.__class__)
                )

            if len(the_bands.shape) == 2 and len(the_labels) != 1:
                raise ValidationError('More array labels than the number of arrays')
            elif len(the_bands.shape) == 3 and len(the_labels) != the_bands.shape[0]:
                raise ValidationError('More array labels than the number of arrays')
        else:
            the_labels = None

        return the_bands, the_occupations, the_labels

    def set_bands(self, bands, units=None, occupations=None, labels=None):
        """
        Set an array of band energies of dimension (nkpoints x nbands).
        Kpoints must be set in advance. Can contain floats or None.
        :param bands: a list of nkpoints lists of nbands bands, or a 2D array
        of shape (nkpoints x nbands), with band energies for each kpoint
        :param units: optional, energy units
        :param occupations: optional, a 2D list or array of floats of same
        shape as bands, with the occupation associated to each band
        """
        # checks bands and occupations
        the_bands, the_occupations, the_labels = self._validate_bands_occupations(bands, occupations, labels)
        # set bands and their units
        self.set_array('bands', the_bands)
        self.units = units

        if the_labels is not None:
            self.base.attributes.set('array_labels', the_labels)

        if the_occupations is not None:
            # set occupations
            self.set_array('occupations', the_occupations)

    @property
    def array_labels(self):
        """
        Get the labels associated with the band arrays
        """
        return self.base.attributes.get('array_labels', None)

    @property
    def units(self):
        """
        Units in which the data in bands were stored. A string
        """
        # return copy.deepcopy(self._pbc)
        return self.base.attributes.get('units')

    @units.setter
    def units(self, value):
        """
        Set the value of pbc, i.e. a tuple of three booleans, indicating if the
        cell is periodic in the 1,2,3 crystal direction
        """
        the_str = str(value)
        self.base.attributes.set('units', the_str)

    def _set_pbc(self, value):
        """
        validate the pbc, then store them
        """
        from aiida.common.exceptions import ModificationNotAllowed
        from aiida.orm.nodes.data.structure import get_valid_pbc

        if self.is_stored:
            raise ModificationNotAllowed('The KpointsData object cannot be modified, it has already been stored')
        the_pbc = get_valid_pbc(value)
        self.base.attributes.set('pbc1', the_pbc[0])
        self.base.attributes.set('pbc2', the_pbc[1])
        self.base.attributes.set('pbc3', the_pbc[2])

    def get_bands(self, also_occupations=False, also_labels=False):
        """
        Returns an array (nkpoints x num_bands or nspins x nkpoints x num_bands)
        of energies.
        :param also_occupations: if True, returns also the occupations array.
        Default = False
        """
        try:
            bands = numpy.array(self.get_array('bands'))
        except KeyError:
            raise AttributeError('No stored bands has been found')

        to_return = [bands]

        if also_occupations:
            try:
                occupations = numpy.array(self.get_array('occupations'))
            except KeyError:
                raise AttributeError('No occupations were set')
            to_return.append(occupations)

        if also_labels:
            to_return.append(self.array_labels)

        if len(to_return) == 1:
            return bands

        return to_return

    def _get_bandplot_data(self, cartesian, prettify_format=None, join_symbol=None, get_segments=False, y_origin=0.):
        """
        Get data to plot a band structure

        :param cartesian: if True, distances (for the x-axis) are computed in
                cartesian coordinates, otherwise they are computed in reciprocal
                coordinates. cartesian=True will fail if no cell has been set.
        :param prettify_format: by default, strings are not prettified. If you want
             to prettify them, pass a valid prettify_format string (see valid options
             in the docstring of :py:func:prettify_labels).
        :param join_symbols: by default, strings are not joined. If you pass a string,
             this is used to join strings that are much closer than a given threshold.
             The most typical string is the pipe symbol: ``|``.
        :param get_segments: if True, also computes the band split into segments
        :param y_origin: if present, shift bands so to set the value specified at ``y=0``
        :return: a plot_info dictiorary, whose keys are ``x`` (array of distances
           for the x axis of the plot); ``y`` (array of bands), ``labels`` (list
           of tuples in the format (float x value of the label, label string),
           ``band_type_idx`` (array containing an index for each band: if there is only
           one spin, then it's an array of zeros, of length equal to the number of bands
           at each point; if there are two spins, then it's an array of zeros or ones
           depending on the type of spin; the length is always equalt to the total
           number of bands per kpoint).
        """
        # pylint: disable=too-many-locals,too-many-branches,too-many-statements
        # load the x and y's of the graph
        stored_bands = self.get_bands()
        if len(stored_bands.shape) == 2:
            bands = stored_bands
            band_type_idx = numpy.array([0] * stored_bands.shape[1])
            two_band_types = False
        elif len(stored_bands.shape) == 3:
            bands = numpy.concatenate(stored_bands, axis=1)
            band_type_idx = numpy.array([0] * stored_bands.shape[2] + [1] * stored_bands.shape[2])
            two_band_types = True
        else:
            raise ValueError('Unexpected shape of bands')

        bands -= y_origin

        # here I build the x distances on the graph (in cartesian coordinates
        # if cartesian==True AND if the cell was set, otherwise in reciprocal
        # coordinates)
        try:
            kpoints = self.get_kpoints(cartesian=cartesian)
        except AttributeError:
            # this error is happening if cartesian==True and if no cell has been
            # set -> we switch to reciprocal coordinates to compute distances
            kpoints = self.get_kpoints()
        # I take advantage of the path to recognize discontinuities
        try:
            labels = self.labels
            labels_indices = [i[0] for i in labels]
        except (AttributeError, TypeError):
            labels = []
            labels_indices = []

        # since I can have discontinuous paths, I set on those points the distance to zero
        # as a result, where there are discontinuities in the path,
        # I have two consecutive points with the same x coordinate
        distances = [
            numpy.linalg.norm(kpoints[i] -
                              kpoints[i - 1]) if not (i in labels_indices and i - 1 in labels_indices) else 0.
            for i in range(1, len(kpoints))
        ]
        x = [float(sum(distances[:i])) for i in range(len(distances) + 1)]

        # transform the index of the labels in the coordinates of x
        raw_labels = [(x[i[0]], i[1]) for i in labels]

        the_labels = raw_labels

        if prettify_format:
            the_labels = prettify_labels(the_labels, format=prettify_format)
        if join_symbol:
            the_labels = join_labels(the_labels, join_symbol=join_symbol)

        plot_info = {}
        plot_info['x'] = x
        plot_info['y'] = bands
        plot_info['band_type_idx'] = band_type_idx
        plot_info['raw_labels'] = raw_labels
        plot_info['labels'] = the_labels

        if get_segments:
            plot_info['path'] = []
            plot_info['paths'] = []

            if len(labels) > 1:
                # I add an empty label that points to the first band if the first label does not do it
                if labels[0][0] != 0:
                    labels.insert(0, (0, ''))
                # I add an empty label that points to the last band if the last label does not do it
                if labels[-1][0] != len(bands) - 1:
                    labels.append((len(bands) - 1, ''))
                for (position_from, label_from), (position_to, label_to) in zip(labels[:-1], labels[1:]):
                    if position_to - position_from > 1:
                        # Create a new path line only if there are at least two points,
                        # otherwise it is probably just a discontinuity point in the band
                        # structure (e.g. Gamma-X|Y-Gamma), where X and Y would be two
                        # consecutive points, but there is no path between them
                        plot_info['path'].append([label_from, label_to])
                    path_dict = {
                        'length': position_to - position_from,
                        'from': label_from,
                        'to': label_to,
                        'values': bands[position_from:position_to + 1, :].transpose().tolist(),
                        'x': x[position_from:position_to + 1],
                        'two_band_types': two_band_types,
                    }
                    plot_info['paths'].append(path_dict)
            else:
                label_from = '0'
                label_to = '1'
                path_dict = {
                    'length': bands.shape[0] - 1,
                    'from': label_from,
                    'to': label_to,
                    'values': bands.transpose().tolist(),
                    'x': x,
                    'two_band_types': two_band_types,
                }
                plot_info['paths'].append(path_dict)
                plot_info['path'].append([label_from, label_to])

        return plot_info

    def _prepare_agr_batch(self, main_file_name='', comments=True, prettify_format=None):
        """
        Prepare two files, data and batch, to be plot with xmgrace as:
        xmgrace -batch file.dat

        :param main_file_name: if the user asks to write the main content on a
             file, this contains the filename. This should be used to infer a
             good filename for the additional files.
             In this case, we remove the extension, and add '_data.dat'
        :param comments: if True, print comments (if it makes sense for the given
            format)
        :param prettify_format: if None, use the default prettify format. Otherwise
            specify a string with the prettifier to use.
        """
        # pylint: disable=too-many-locals
        import os

        dat_filename = os.path.splitext(main_file_name)[0] + '_data.dat'

        if prettify_format is None:
            # Default. Specified like this to allow caller functions to pass 'None'
            prettify_format = 'agr_seekpath'

        plot_info = self._get_bandplot_data(cartesian=True, prettify_format=prettify_format, join_symbol='|')

        bands = plot_info['y']
        x = plot_info['x']
        labels = plot_info['labels']

        num_bands = bands.shape[1]

        # axis limits
        y_max_lim = bands.max()
        y_min_lim = bands.min()
        x_min_lim = min(x)  # this isn't a numpy array, but a list
        x_max_lim = max(x)

        # first prepare the xy coordinates of the sets
        raw_data, _ = self._prepare_dat_blocks(plot_info)

        batch = []
        if comments:
            batch.append(prepare_header_comment(self.uuid, plot_info, comment_char='#'))

        batch.append(f'READ XY "{dat_filename}"')

        # axis limits
        batch.append(f'world {x_min_lim}, {y_min_lim}, {x_max_lim}, {y_max_lim}')

        # axis label
        batch.append('yaxis label "Dispersion"')

        # axis ticks
        batch.append('xaxis  tick place both')
        batch.append('xaxis  tick spec type both')
        batch.append(f'xaxis  tick spec {len(labels)}')
        # set the name of the special points
        for index, label in enumerate(labels):
            batch.append(f'xaxis  tick major {index}, {label[0]}')
            batch.append(f'xaxis  ticklabel {index}, "{label[1]}"')
        batch.append('xaxis  tick major color 7')
        batch.append('xaxis  tick major grid on')

        # minor graphical tweak
        batch.append('yaxis  tick minor ticks 3')
        batch.append('frame linewidth 1.0')

        # use helvetica fonts
        batch.append('map font 4 to "Helvetica", "Helvetica"')
        batch.append('yaxis  label font 4')
        batch.append('xaxis  label font 4')

        # set color and linewidths of bands
        for index in range(num_bands):
            batch.append(f's{index} line color 1')
            batch.append(f's{index} linewidth 1')

        batch_data = '\n'.join(batch) + '\n'
        extra_files = {dat_filename: raw_data}

        return batch_data.encode('utf-8'), extra_files

    def _prepare_dat_multicolumn(self, main_file_name='', comments=True):  # pylint: disable=unused-argument
        """
        Write an N x M matrix. First column is the distance between kpoints,
        The other columns are the bands. Header contains number of kpoints and
        the number of bands (commented).

        :param comments: if True, print comments (if it makes sense for the given
            format)
        """
        plot_info = self._get_bandplot_data(cartesian=True, prettify_format=None, join_symbol='|')

        bands = plot_info['y']
        x = plot_info['x']

        return_text = []
        if comments:
            return_text.append(prepare_header_comment(self.uuid, plot_info, comment_char='#'))

        for i in zip(x, bands):
            line = [f'{i[0]:.8f}'] + [f'{j:.8f}' for j in i[1]]
            return_text.append('\t'.join(line))

        return ('\n'.join(return_text) + '\n').encode('utf-8'), {}

    def _prepare_dat_blocks(self, main_file_name='', comments=True):  # pylint: disable=unused-argument
        """
        Format suitable for gnuplot using blocks.
        Columns with x and y (path and band energy). Several blocks, separated
        by two empty lines, one per energy band.

        :param comments: if True, print comments (if it makes sense for the given
            format)
        """
        plot_info = self._get_bandplot_data(cartesian=True, prettify_format=None, join_symbol='|')

        bands = plot_info['y']
        x = plot_info['x']

        return_text = []
        if comments:
            return_text.append(prepare_header_comment(self.uuid, plot_info, comment_char='#'))

        for band in numpy.transpose(bands):
            for i in zip(x, band):
                line = [f'{i[0]:.8f}', f'{i[1]:.8f}']
                return_text.append('\t'.join(line))
            return_text.append('')
            return_text.append('')

        return '\n'.join(return_text).encode('utf-8'), {}

    def _matplotlib_get_dict(
        self,
        main_file_name='',
        comments=True,
        title='',
        legend=None,
        legend2=None,
        y_max_lim=None,
        y_min_lim=None,
        y_origin=0.,
        prettify_format=None,
        **kwargs
    ):  # pylint: disable=unused-argument
        """
        Prepare the data to send to the python-matplotlib plotting script.

        :param comments: if True, print comments (if it makes sense for the given
            format)
        :param plot_info: a dictionary
        :param setnumber_offset: an offset to be applied to all set numbers
            (i.e. s0 is replaced by s[offset], s1 by s[offset+1], etc.)
        :param color_number: the color number for lines, symbols, error bars
            and filling (should be less than the parameter MAX_NUM_AGR_COLORS
            defined below)
        :param title: the title
        :param legend: the legend (applied only to the first of the set)
        :param legend2: the legend for second-type spins
            (applied only to the first of the set)
        :param y_max_lim: the maximum on the y axis (if None, put the
            maximum of the bands)
        :param y_min_lim: the minimum on the y axis (if None, put the
            minimum of the bands)
        :param y_origin: the new origin of the y axis -> all bands are replaced
            by bands-y_origin
        :param prettify_format: if None, use the default prettify format. Otherwise
            specify a string with the prettifier to use.
        :param kwargs: additional customization variables; only a subset is
            accepted, see internal variable 'valid_additional_keywords
        """
        # pylint: disable=too-many-arguments,too-many-locals

        # Only these keywords are accepted in kwargs, and then set into the json
        valid_additional_keywords = [
            'bands_color',  # Color of band lines
            'bands_linewidth',  # linewidth of bands
            'bands_linestyle',  # linestyle of bands
            'bands_marker',  # marker for bands
            'bands_markersize',  # size of the marker of bands
            'bands_markeredgecolor',  # marker edge color for bands
            'bands_markeredgewidth',  # marker edge width for bands
            'bands_markerfacecolor',  # marker face color for bands
            'bands_color2',  # Color of band lines (for other spin, if present)
            'bands_linewidth2',  # linewidth of bands (for other spin, if present)
            'bands_linestyle2',  # linestyle of bands (for other spin, if present)
            'bands_marker2',  # marker for bands (for other spin, if present)
            'bands_markersize2',  # size of the marker of bands (for other spin, if present)
            'bands_markeredgecolor2',  # marker edge color for bands (for other spin, if present)
            'bands_markeredgewidth2',  # marker edge width for bands (for other spin, if present)
            'bands_markerfacecolor2',  # marker face color for bands (for other spin, if present)
            'plot_zero_axis',  # If true, plot an axis at y=0
            'zero_axis_color',  # Color of the axis at y=0
            'zero_axis_linestyle',  # linestyle of the axis at y=0
            'zero_axis_linewidth',  # linewidth of the axis at y=0
            'use_latex',  # If true, use latex to render captions
        ]

        # Note: I do not want to import matplotlib here, for two reasons:
        # 1. I would like to be able to print the script for the user
        # 2. I don't want to mess up with the user matplotlib backend
        #    (that I should do if the user does not have a X server, but that
        #    I do not want to do if he's e.g. in jupyter)
        # Therefore I just create a string that can be executed as needed, e.g. with eval.
        # I take care of sanitizing the output.
        if prettify_format is None:
            # Default. Specified like this to allow caller functions to pass 'None'
            prettify_format = 'latex_seekpath'

        # The default for use_latex is False
        join_symbol = r'\textbar{}' if kwargs.get('use_latex', False) else '|'

        plot_info = self._get_bandplot_data(
            cartesian=True,
            prettify_format=prettify_format,
            join_symbol=join_symbol,
            get_segments=True,
            y_origin=y_origin
        )

        all_data = {}

        bands = plot_info['y']
        x = plot_info['x']
        labels = plot_info['labels']
        # prepare xticks labels
        if labels:
            tick_pos, tick_labels = zip(*labels)
        else:
            tick_pos = []
            tick_labels = []

        all_data['paths'] = plot_info['paths']
        all_data['band_type_idx'] = plot_info['band_type_idx'].tolist()

        all_data['tick_pos'] = tick_pos
        all_data['tick_labels'] = tick_labels
        all_data['legend_text'] = legend
        all_data['legend_text2'] = legend2
        all_data['yaxis_label'] = f'Dispersion ({self.units})'
        all_data['title'] = title
        if comments:
            all_data['comment'] = prepare_header_comment(self.uuid, plot_info, comment_char='#')

        # axis limits
        if y_max_lim is None:
            y_max_lim = numpy.nanmax(bands)
        if y_min_lim is None:
            y_min_lim = numpy.nanmin(bands)
        x_min_lim = min(x)  # this isn't a numpy array, but a list
        x_max_lim = max(x)
        all_data['x_min_lim'] = x_min_lim
        all_data['x_max_lim'] = x_max_lim
        all_data['y_min_lim'] = y_min_lim
        all_data['y_max_lim'] = y_max_lim

        for key, value in kwargs.items():
            if key not in valid_additional_keywords:
                raise TypeError(f"_matplotlib_get_dict() got an unexpected keyword argument '{key}'")
            all_data[key] = value

        return all_data

    def _prepare_mpl_singlefile(self, *args, **kwargs):
        """
        Prepare a python script using matplotlib to plot the bands

        For the possible parameters, see documentation of
        :py:meth:`~aiida.orm.nodes.data.array.bands.BandsData._matplotlib_get_dict`
        """
        all_data = self._matplotlib_get_dict(*args, **kwargs)

        s_header = MATPLOTLIB_HEADER_TEMPLATE.substitute()
        s_import = MATPLOTLIB_IMPORT_DATA_INLINE_TEMPLATE.substitute(all_data_json=json.dumps(all_data, indent=2))
        s_body = self._get_mpl_body_template(all_data['paths'])
        s_footer = MATPLOTLIB_FOOTER_TEMPLATE_SHOW.substitute()

        string = s_header + s_import + s_body + s_footer

        return string.encode('utf-8'), {}

    def _prepare_mpl_withjson(self, main_file_name='', *args, **kwargs):  # pylint: disable=keyword-arg-before-vararg
        """
        Prepare a python script using matplotlib to plot the bands, with the JSON
        returned as an independent file.

        For the possible parameters, see documentation of
        :py:meth:`~aiida.orm.nodes.data.array.bands.BandsData._matplotlib_get_dict`
        """
        import os

        all_data = self._matplotlib_get_dict(*args, main_file_name=main_file_name, **kwargs)

        json_fname = os.path.splitext(main_file_name)[0] + '_data.json'
        # Escape double_quotes
        json_fname = json_fname.replace('"', '\"')

        ext_files = {json_fname: json.dumps(all_data, indent=2).encode('utf-8')}

        s_header = MATPLOTLIB_HEADER_TEMPLATE.substitute()
        s_import = MATPLOTLIB_IMPORT_DATA_FROMFILE_TEMPLATE.substitute(json_fname=json_fname)
        s_body = self._get_mpl_body_template(all_data['paths'])
        s_footer = MATPLOTLIB_FOOTER_TEMPLATE_SHOW.substitute()

        string = s_header + s_import + s_body + s_footer

        return string.encode('utf-8'), ext_files

    def _prepare_mpl_pdf(self, main_file_name='', *args, **kwargs):  # pylint: disable=keyword-arg-before-vararg,unused-argument
        """
        Prepare a python script using matplotlib to plot the bands, with the JSON
        returned as an independent file.

        For the possible parameters, see documentation of
        :py:meth:`~aiida.orm.nodes.data.array.bands.BandsData._matplotlib_get_dict`
        """
        import os
        import subprocess
        import sys
        import tempfile

        all_data = self._matplotlib_get_dict(*args, **kwargs)

        # Use the Agg backend
        s_header = MATPLOTLIB_HEADER_AGG_TEMPLATE.substitute()
        s_import = MATPLOTLIB_IMPORT_DATA_INLINE_TEMPLATE.substitute(all_data_json=json.dumps(all_data, indent=2))
        s_body = self._get_mpl_body_template(all_data['paths'])

        # I get a temporary file name
        handle, filename = tempfile.mkstemp()
        os.close(handle)
        os.remove(filename)

        escaped_fname = filename.replace('"', '\"')

        s_footer = MATPLOTLIB_FOOTER_TEMPLATE_EXPORTFILE.substitute(fname=escaped_fname, format='pdf')

        string = s_header + s_import + s_body + s_footer

        # I don't exec it because I might mess up with the matplotlib backend etc.
        # I run instead in a different process, with the same executable
        # (so it should work properly with virtualenvs)
        with tempfile.NamedTemporaryFile(mode='w+') as handle:
            handle.write(string)
            handle.flush()
            subprocess.check_output([sys.executable, handle.name])

        if not os.path.exists(filename):
            raise RuntimeError('Unable to generate the PDF...')

        with open(filename, 'rb', encoding=None) as handle:
            imgdata = handle.read()
        os.remove(filename)

        return imgdata, {}

    def _prepare_mpl_png(self, main_file_name='', *args, **kwargs):  # pylint: disable=keyword-arg-before-vararg,unused-argument
        """
        Prepare a python script using matplotlib to plot the bands, with the JSON
        returned as an independent file.

        For the possible parameters, see documentation of
        :py:meth:`~aiida.orm.nodes.data.array.bands.BandsData._matplotlib_get_dict`
        """
        import os
        import subprocess
        import sys
        import tempfile

        all_data = self._matplotlib_get_dict(*args, **kwargs)

        # Use the Agg backend
        s_header = MATPLOTLIB_HEADER_AGG_TEMPLATE.substitute()
        s_import = MATPLOTLIB_IMPORT_DATA_INLINE_TEMPLATE.substitute(all_data_json=json.dumps(all_data, indent=2))
        s_body = self._get_mpl_body_template(all_data['paths'])

        # I get a temporary file name
        handle, filename = tempfile.mkstemp()
        os.close(handle)
        os.remove(filename)

        escaped_fname = filename.replace('"', '\"')

        s_footer = MATPLOTLIB_FOOTER_TEMPLATE_EXPORTFILE_WITH_DPI.substitute(fname=escaped_fname, format='png', dpi=300)

        string = s_header + s_import + s_body + s_footer

        # I don't exec it because I might mess up with the matplotlib backend etc.
        # I run instead in a different process, with the same executable
        # (so it should work properly with virtualenvs)
        with tempfile.NamedTemporaryFile(mode='w+') as handle:
            handle.write(string)
            handle.flush()
            subprocess.check_output([sys.executable, handle.name])

        if not os.path.exists(filename):
            raise RuntimeError('Unable to generate the PNG...')

        with open(filename, 'rb', encoding=None) as handle:
            imgdata = handle.read()
        os.remove(filename)

        return imgdata, {}

    @staticmethod
    def _get_mpl_body_template(paths):
        """
        :param paths: paths of k-points
        """
        if len(paths) == 1:
            s_body = MATPLOTLIB_BODY_TEMPLATE.substitute(plot_code=SINGLE_KP)
        else:
            s_body = MATPLOTLIB_BODY_TEMPLATE.substitute(plot_code=MULTI_KP)
        return s_body

    def show_mpl(self, **kwargs):
        """
        Call a show() command for the band structure using matplotlib.
        This uses internally the 'mpl_singlefile' format, with empty
        main_file_name.

        Other kwargs are passed to self._exportcontent.
        """
        exec(*self._exportcontent(fileformat='mpl_singlefile', main_file_name='', **kwargs))  # pylint: disable=exec-used

    def _prepare_gnuplot(
        self,
        main_file_name=None,
        title='',
        comments=True,
        prettify_format=None,
        y_max_lim=None,
        y_min_lim=None,
        y_origin=0.
    ):
        """
        Prepare an gnuplot script to plot the bands, with the .dat file
        returned as an independent file.

        :param main_file_name: if the user asks to write the main content on a
             file, this contains the filename. This should be used to infer a
             good filename for the additional files.
             In this case, we remove the extension, and add '_data.dat'
        :param title: if specified, add a title to the plot
        :param comments: if True, print comments (if it makes sense for the given
            format)
        :param prettify_format: if None, use the default prettify format. Otherwise
            specify a string with the prettifier to use.
        """
        # pylint: disable=too-many-arguments,too-many-locals
        import os

        main_file_name = main_file_name or 'band.dat'
        dat_filename = os.path.splitext(main_file_name)[0] + '_data.dat'

        if prettify_format is None:
            # Default. Specified like this to allow caller functions to pass 'None'
            prettify_format = 'gnuplot_seekpath'

        plot_info = self._get_bandplot_data(
            cartesian=True, prettify_format=prettify_format, join_symbol='|', y_origin=y_origin
        )

        bands = plot_info['y']
        x = plot_info['x']

        # axis limits
        if y_max_lim is None:
            y_max_lim = bands.max()
        if y_min_lim is None:
            y_min_lim = bands.min()
        x_min_lim = min(x)  # this isn't a numpy array, but a list
        x_max_lim = max(x)

        # first prepare the xy coordinates of the sets
        raw_data, _ = self._prepare_dat_blocks(plot_info, comments=comments)

        xtics_string = ', '.join(f'"{label}" {pos}' for pos, label in plot_info['labels'])

        script = []
        # Start with some useful comments

        if comments:
            script.append(prepare_header_comment(self.uuid, plot_info=plot_info, comment_char='# '))
        script.append('')

        script.append(
            """## Uncomment the next two lines to write directly to PDF
## Note: You need to have gnuplot installed with pdfcairo support!
#set term pdfcairo
#set output 'out.pdf'

### Uncomment one of the options below to change font
### For the LaTeX fonts, you can download them from here:
### https://sourceforge.net/projects/cm-unicode/
### And then install them in your system
## LaTeX Serif font, if installed
#set termopt font "CMU Serif, 12"
## LaTeX Sans Serif font, if installed
#set termopt font "CMU Sans Serif, 12"
## Classical Times New Roman
#set termopt font "Times New Roman, 12"
"""
        )

        # Actual logic
        script.append('set termopt enhanced')  # Properly deals with e.g. subscripts
        script.append('set encoding utf8')  # To deal with Greek letters
        script.append(f'set xtics ({xtics_string})')
        script.append('unset key')
        script.append(f'set yrange [{y_min_lim}:{y_max_lim}]')
        script.append(f"set ylabel \"Dispersion ({self.units})\"")

        if title:
            script.append('set title "{}"'.format(title.replace('"', '\"')))

        # Plot, escaping filename
        if len(x) > 1:
            script.append(f'set xrange [{x_min_lim}:{x_max_lim}]')
            script.append('set grid xtics lt 1 lc rgb "#888888"')
            script.append('plot "{}" with l lc rgb "#000000"'.format(os.path.basename(dat_filename).replace('"', '\"')))
        else:
            script.append('set xrange [-1.0:1.0]')
            script.append(
                'plot "{}" using ($1-0.25):($2):(0.5):(0) with vectors nohead lc rgb "#000000"'.format(
                    os.path.basename(dat_filename).replace('"', '\"')
                )
            )

        script_data = '\n'.join(script) + '\n'
        extra_files = {dat_filename: raw_data}

        return script_data.encode('utf-8'), extra_files

    def _prepare_agr(
        self,
        main_file_name='',
        comments=True,
        setnumber_offset=0,
        color_number=1,
        color_number2=2,
        legend='',
        title='',
        y_max_lim=None,
        y_min_lim=None,
        y_origin=0.,
        prettify_format=None
    ):
        """
        Prepare an xmgrace agr file.

        :param comments: if True, print comments
            (if it makes sense for the given format)
        :param plot_info: a dictionary
        :param setnumber_offset: an offset to be applied to all set numbers
            (i.e. s0 is replaced by s[offset], s1 by s[offset+1], etc.)
        :param color_number: the color number for lines, symbols, error bars
            and filling (should be less than the parameter MAX_NUM_AGR_COLORS
            defined below)
        :param color_number2: the color number for lines, symbols, error bars
            and filling for the second-type spins (should be less than the
            parameter MAX_NUM_AGR_COLORS defined below)
        :param legend: the legend (applied only to the first set)
        :param title: the title
        :param y_max_lim: the maximum on the y axis (if None, put the
            maximum of the bands); applied *after* shifting the origin
            by ``y_origin``
        :param y_min_lim: the minimum on the y axis (if None, put the
            minimum of the bands); applied *after* shifting the origin
            by ``y_origin``
        :param y_origin: the new origin of the y axis -> all bands are replaced
            by bands-y_origin
        :param prettify_format: if None, use the default prettify format. Otherwise
            specify a string with the prettifier to use.
        """
        # pylint: disable=too-many-arguments,too-many-locals,too-many-branches,unused-argument
        if prettify_format is None:
            # Default. Specified like this to allow caller functions to pass 'None'
            prettify_format = 'agr_seekpath'

        plot_info = self._get_bandplot_data(
            cartesian=True, prettify_format=prettify_format, join_symbol='|', y_origin=y_origin
        )

        import math

        # load the x and y of every set
        if color_number > MAX_NUM_AGR_COLORS:
            raise ValueError(f'Color number is too high (should be less than {MAX_NUM_AGR_COLORS})')
        if color_number2 > MAX_NUM_AGR_COLORS:
            raise ValueError(f'Color number 2 is too high (should be less than {MAX_NUM_AGR_COLORS})')

        bands = plot_info['y']
        x = plot_info['x']
        the_bands = numpy.transpose(bands)
        labels = plot_info['labels']
        num_labels = len(labels)

        # axis limits
        if y_max_lim is None:
            y_max_lim = the_bands.max()
        if y_min_lim is None:
            y_min_lim = the_bands.min()
        x_min_lim = min(x)  # this isn't a numpy array, but a list
        x_max_lim = max(x)
        ytick_spacing = 10**int(math.log10((y_max_lim - y_min_lim)))

        # prepare xticks labels
        sx1 = ''
        for i, label in enumerate(labels):
            sx1 += AGR_SINGLE_XTICK_TEMPLATE.substitute(
                index=i,
                coord=label[0],
                name=label[1],
            )
        xticks = AGR_XTICKS_TEMPLATE.substitute(
            num_labels=num_labels,
            single_xtick_templates=sx1,
        )

        # build the arrays with the xy coordinates
        all_sets = []
        for band in the_bands:
            this_set = ''
            for i in zip(x, band):
                line = f'{i[0]:.8f}' + '\t' + f'{i[1]:.8f}' + '\n'
                this_set += line
            all_sets.append(this_set)

        set_descriptions = ''
        for i, (this_set, band_type) in enumerate(zip(all_sets, plot_info['band_type_idx'])):
            if band_type % 2 == 0:
                linecolor = color_number
            else:
                linecolor = color_number2
            width = str(2.0)
            set_descriptions += AGR_SET_DESCRIPTION_TEMPLATE.substitute(
                set_number=i + setnumber_offset,
                linewidth=width,
                color_number=linecolor,
                legend=legend if i == 0 else ''
            )

        units = self.units

        graphs = AGR_GRAPH_TEMPLATE.substitute(
            x_min_lim=x_min_lim,
            y_min_lim=y_min_lim,
            x_max_lim=x_max_lim,
            y_max_lim=y_max_lim,
            yaxislabel=f'Dispersion ({units})',
            xticks_template=xticks,
            set_descriptions=set_descriptions,
            ytick_spacing=ytick_spacing,
            title=title,
        )
        sets = []
        for i, this_set in enumerate(all_sets):
            sets.append(AGR_SINGLESET_TEMPLATE.substitute(set_number=i + setnumber_offset, xydata=this_set))
        the_sets = '&\n'.join(sets)

        string = AGR_TEMPLATE.substitute(graphs=graphs, sets=the_sets)

        if comments:
            string = prepare_header_comment(self.uuid, plot_info, comment_char='#') + '\n' + string

        return string.encode('utf-8'), {}

    def _get_band_segments(self, cartesian):
        """Return the band segments."""
        plot_info = self._get_bandplot_data(
            cartesian=cartesian, prettify_format=None, join_symbol=None, get_segments=True
        )

        out_dict = {'label': self.label}

        out_dict['path'] = plot_info['path']
        out_dict['paths'] = plot_info['paths']

        return out_dict

    def _prepare_json(self, main_file_name='', comments=True):  # pylint: disable=unused-argument
        """
        Prepare a json file in a format compatible with the AiiDA band visualizer

        :param comments: if True, print comments (if it makes sense for the given
            format)
        """
        from aiida import get_file_header

        json_dict = self._get_band_segments(cartesian=True)
        json_dict['original_uuid'] = self.uuid

        if comments:
            json_dict['comments'] = get_file_header(comment_char='')

        return json.dumps(json_dict).encode('utf-8'), {}


MAX_NUM_AGR_COLORS = 15

AGR_TEMPLATE = Template(
    """
    # Grace project file
    #
    @version 50122
    @page size 792, 612
    @page scroll 5%
    @page inout 5%
    @link page off
    @map font 8 to "Courier", "Courier"
    @map font 10 to "Courier-Bold", "Courier-Bold"
    @map font 11 to "Courier-BoldOblique", "Courier-BoldOblique"
    @map font 9 to "Courier-Oblique", "Courier-Oblique"
    @map font 4 to "Helvetica", "Helvetica"
    @map font 6 to "Helvetica-Bold", "Helvetica-Bold"
    @map font 7 to "Helvetica-BoldOblique", "Helvetica-BoldOblique"
    @map font 5 to "Helvetica-Oblique", "Helvetica-Oblique"
    @map font 14 to "NimbusMonoL-BoldOblique", "NimbusMonoL-BoldOblique"
    @map font 15 to "NimbusMonoL-Regular", "NimbusMonoL-Regular"
    @map font 16 to "NimbusMonoL-RegularOblique", "NimbusMonoL-RegularOblique"
    @map font 17 to "NimbusRomanNo9L-Medium", "NimbusRomanNo9L-Medium"
    @map font 18 to "NimbusRomanNo9L-MediumItalic", "NimbusRomanNo9L-MediumItalic"
    @map font 19 to "NimbusRomanNo9L-Regular", "NimbusRomanNo9L-Regular"
    @map font 20 to "NimbusRomanNo9L-RegularItalic", "NimbusRomanNo9L-RegularItalic"
    @map font 21 to "NimbusSansL-Bold", "NimbusSansL-Bold"
    @map font 22 to "NimbusSansL-BoldCondensed", "NimbusSansL-BoldCondensed"
    @map font 23 to "NimbusSansL-BoldCondensedItalic", "NimbusSansL-BoldCondensedItalic"
    @map font 24 to "NimbusSansL-BoldItalic", "NimbusSansL-BoldItalic"
    @map font 25 to "NimbusSansL-Regular", "NimbusSansL-Regular"
    @map font 26 to "NimbusSansL-RegularCondensed", "NimbusSansL-RegularCondensed"
    @map font 27 to "NimbusSansL-RegularCondensedItalic", "NimbusSansL-RegularCondensedItalic"
    @map font 28 to "NimbusSansL-RegularItalic", "NimbusSansL-RegularItalic"
    @map font 29 to "StandardSymbolsL-Regular", "StandardSymbolsL-Regular"
    @map font 12 to "Symbol", "Symbol"
    @map font 31 to "Symbol-Regular", "Symbol-Regular"
    @map font 2 to "Times-Bold", "Times-Bold"
    @map font 3 to "Times-BoldItalic", "Times-BoldItalic"
    @map font 1 to "Times-Italic", "Times-Italic"
    @map font 0 to "Times-Roman", "Times-Roman"
    @map font 36 to "URWBookmanL-DemiBold", "URWBookmanL-DemiBold"
    @map font 37 to "URWBookmanL-DemiBoldItalic", "URWBookmanL-DemiBoldItalic"
    @map font 38 to "URWBookmanL-Light", "URWBookmanL-Light"
    @map font 39 to "URWBookmanL-LightItalic", "URWBookmanL-LightItalic"
    @map font 40 to "URWChanceryL-MediumItalic", "URWChanceryL-MediumItalic"
    @map font 41 to "URWGothicL-Book", "URWGothicL-Book"
    @map font 42 to "URWGothicL-BookOblique", "URWGothicL-BookOblique"
    @map font 43 to "URWGothicL-Demi", "URWGothicL-Demi"
    @map font 44 to "URWGothicL-DemiOblique", "URWGothicL-DemiOblique"
    @map font 45 to "URWPalladioL-Bold", "URWPalladioL-Bold"
    @map font 46 to "URWPalladioL-BoldItalic", "URWPalladioL-BoldItalic"
    @map font 47 to "URWPalladioL-Italic", "URWPalladioL-Italic"
    @map font 48 to "URWPalladioL-Roman", "URWPalladioL-Roman"
    @map font 13 to "ZapfDingbats", "ZapfDingbats"
    @map color 0 to (255, 255, 255), "white"
    @map color 1 to (0, 0, 0), "black"
    @map color 2 to (255, 0, 0), "red"
    @map color 3 to (0, 255, 0), "green"
    @map color 4 to (0, 0, 255), "blue"
    @map color 5 to (255, 215, 0), "yellow"
    @map color 6 to (188, 143, 143), "brown"
    @map color 7 to (220, 220, 220), "grey"
    @map color 8 to (148, 0, 211), "violet"
    @map color 9 to (0, 255, 255), "cyan"
    @map color 10 to (255, 0, 255), "magenta"
    @map color 11 to (255, 165, 0), "orange"
    @map color 12 to (114, 33, 188), "indigo"
    @map color 13 to (103, 7, 72), "maroon"
    @map color 14 to (64, 224, 208), "turquoise"
    @map color 15 to (0, 139, 0), "green4"
    @reference date 0
    @date wrap off
    @date wrap year 1950
    @default linewidth 1.0
    @default linestyle 1
    @default color 1
    @default pattern 1
    @default font 0
    @default char size 1.000000
    @default symbol size 1.000000
    @default sformat "%.8g"
    @background color 0
    @page background fill on
    @timestamp off
    @timestamp 0.03, 0.03
    @timestamp color 1
    @timestamp rot 0
    @timestamp font 0
    @timestamp char size 1.000000
    @timestamp def "Wed Jul 30 16:44:34 2014"
    @r0 off
    @link r0 to g0
    @r0 type above
    @r0 linestyle 1
    @r0 linewidth 1.0
    @r0 color 1
    @r0 line 0, 0, 0, 0
    @r1 off
    @link r1 to g0
    @r1 type above
    @r1 linestyle 1
    @r1 linewidth 1.0
    @r1 color 1
    @r1 line 0, 0, 0, 0
    @r2 off
    @link r2 to g0
    @r2 type above
    @r2 linestyle 1
    @r2 linewidth 1.0
    @r2 color 1
    @r2 line 0, 0, 0, 0
    @r3 off
    @link r3 to g0
    @r3 type above
    @r3 linestyle 1
    @r3 linewidth 1.0
    @r3 color 1
    @r3 line 0, 0, 0, 0
    @r4 off
    @link r4 to g0
    @r4 type above
    @r4 linestyle 1
    @r4 linewidth 1.0
    @r4 color 1
    @r4 line 0, 0, 0, 0
    $graphs
    $sets
    """
)

AGR_XTICKS_TEMPLATE = Template("""
    @    xaxis  tick spec $num_labels
    $single_xtick_templates
    """)

AGR_SINGLE_XTICK_TEMPLATE = Template(
    """
    @    xaxis  tick major $index, $coord
    @    xaxis  ticklabel $index, "$name"
    """
)

AGR_GRAPH_TEMPLATE = Template(
    """
    @g0 on
    @g0 hidden false
    @g0 type XY
    @g0 stacked false
    @g0 bar hgap 0.000000
    @g0 fixedpoint off
    @g0 fixedpoint type 0
    @g0 fixedpoint xy 0.000000, 0.000000
    @g0 fixedpoint format general general
    @g0 fixedpoint prec 6, 6
    @with g0
    @    world $x_min_lim, $y_min_lim, $x_max_lim, $y_max_lim
    @    stack world 0, 0, 0, 0
    @    znorm 1
    @    view 0.150000, 0.150000, 1.150000, 0.850000
    @    title "$title"
    @    title font 0
    @    title size 1.500000
    @    title color 1
    @    subtitle ""
    @    subtitle font 0
    @    subtitle size 1.000000
    @    subtitle color 1
    @    xaxes scale Normal
    @    yaxes scale Normal
    @    xaxes invert off
    @    yaxes invert off
    @    xaxis  on
    @    xaxis  type zero false
    @    xaxis  offset 0.000000 , 0.000000
    @    xaxis  bar on
    @    xaxis  bar color 1
    @    xaxis  bar linestyle 1
    @    xaxis  bar linewidth 1.0
    @    xaxis  label ""
    @    xaxis  label layout para
    @    xaxis  label place auto
    @    xaxis  label char size 1.000000
    @    xaxis  label font 4
    @    xaxis  label color 1
    @    xaxis  label place normal
    @    xaxis  tick on
    @    xaxis  tick major 5
    @    xaxis  tick minor ticks 0
    @    xaxis  tick default 6
    @    xaxis  tick place rounded true
    @    xaxis  tick in
    @    xaxis  tick major size 1.000000
    @    xaxis  tick major color 1
    @    xaxis  tick major linewidth 1.0
    @    xaxis  tick major linestyle 1
    @    xaxis  tick major grid on
    @    xaxis  tick minor color 1
    @    xaxis  tick minor linewidth 1.0
    @    xaxis  tick minor linestyle 1
    @    xaxis  tick minor grid off
    @    xaxis  tick minor size 0.500000
    @    xaxis  ticklabel on
    @    xaxis  ticklabel format general
    @    xaxis  ticklabel prec 5
    @    xaxis  ticklabel formula ""
    @    xaxis  ticklabel append ""
    @    xaxis  ticklabel prepend ""
    @    xaxis  ticklabel angle 0
    @    xaxis  ticklabel skip 0
    @    xaxis  ticklabel stagger 0
    @    xaxis  ticklabel place normal
    @    xaxis  ticklabel offset auto
    @    xaxis  ticklabel offset 0.000000 , 0.010000
    @    xaxis  ticklabel start type auto
    @    xaxis  ticklabel start 0.000000
    @    xaxis  ticklabel stop type auto
    @    xaxis  ticklabel stop 0.000000
    @    xaxis  ticklabel char size 1.500000
    @    xaxis  ticklabel font 4
    @    xaxis  ticklabel color 1
    @    xaxis  tick place both
    @    xaxis  tick spec type both
    $xticks_template
    @    yaxis  on
    @    yaxis  type zero false
    @    yaxis  offset 0.000000 , 0.000000
    @    yaxis  bar on
    @    yaxis  bar color 1
    @    yaxis  bar linestyle 1
    @    yaxis  bar linewidth 1.0
    @    yaxis  label "$yaxislabel"
    @    yaxis  label layout para
    @    yaxis  label place auto
    @    yaxis  label char size 1.500000
    @    yaxis  label font 4
    @    yaxis  label color 1
    @    yaxis  label place normal
    @    yaxis  tick on
    @    yaxis  tick major $ytick_spacing
    @    yaxis  tick minor ticks 1
    @    yaxis  tick default 6
    @    yaxis  tick place rounded true
    @    yaxis  tick in
    @    yaxis  tick major size 1.000000
    @    yaxis  tick major color 1
    @    yaxis  tick major linewidth 1.0
    @    yaxis  tick major linestyle 1
    @    yaxis  tick major grid off
    @    yaxis  tick minor color 1
    @    yaxis  tick minor linewidth 1.0
    @    yaxis  tick minor linestyle 1
    @    yaxis  tick minor grid off
    @    yaxis  tick minor size 0.500000
    @    yaxis  ticklabel on
    @    yaxis  ticklabel format general
    @    yaxis  ticklabel prec 5
    @    yaxis  ticklabel formula ""
    @    yaxis  ticklabel append ""
    @    yaxis  ticklabel prepend ""
    @    yaxis  ticklabel angle 0
    @    yaxis  ticklabel skip 0
    @    yaxis  ticklabel stagger 0
    @    yaxis  ticklabel place normal
    @    yaxis  ticklabel offset auto
    @    yaxis  ticklabel offset 0.000000 , 0.010000
    @    yaxis  ticklabel start type auto
    @    yaxis  ticklabel start 0.000000
    @    yaxis  ticklabel stop type auto
    @    yaxis  ticklabel stop 0.000000
    @    yaxis  ticklabel char size 1.250000
    @    yaxis  ticklabel font 4
    @    yaxis  ticklabel color 1
    @    yaxis  tick place both
    @    yaxis  tick spec type none
    @    altxaxis  off
    @    altyaxis  off
    @    legend on
    @    legend loctype view
    @    legend 0.85, 0.8
    @    legend box color 1
    @    legend box pattern 1
    @    legend box linewidth 1.0
    @    legend box linestyle 1
    @    legend box fill color 0
    @    legend box fill pattern 1
    @    legend font 0
    @    legend char size 1.000000
    @    legend color 1
    @    legend length 4
    @    legend vgap 1
    @    legend hgap 1
    @    legend invert false
    @    frame type 0
    @    frame linestyle 1
    @    frame linewidth 1.0
    @    frame color 1
    @    frame pattern 1
    @    frame background color 0
    @    frame background pattern 0
    $set_descriptions
    """
)

AGR_SET_DESCRIPTION_TEMPLATE = Template(
    """
    @    s$set_number hidden false
    @    s$set_number type xy
    @    s$set_number symbol 0
    @    s$set_number symbol size 1.000000
    @    s$set_number symbol color $color_number
    @    s$set_number symbol pattern 1
    @    s$set_number symbol fill color $color_number
    @    s$set_number symbol fill pattern 0
    @    s$set_number symbol linewidth 1.0
    @    s$set_number symbol linestyle 1
    @    s$set_number symbol char 65
    @    s$set_number symbol char font 0
    @    s$set_number symbol skip 0
    @    s$set_number line type 1
    @    s$set_number line linestyle 1
    @    s$set_number line linewidth $linewidth
    @    s$set_number line color $color_number
    @    s$set_number line pattern 1
    @    s$set_number baseline type 0
    @    s$set_number baseline off
    @    s$set_number dropline off
    @    s$set_number fill type 0
    @    s$set_number fill rule 0
    @    s$set_number fill color $color_number
    @    s$set_number fill pattern 1
    @    s$set_number avalue off
    @    s$set_number avalue type 2
    @    s$set_number avalue char size 1.000000
    @    s$set_number avalue font 0
    @    s$set_number avalue color 1
    @    s$set_number avalue rot 0
    @    s$set_number avalue format general
    @    s$set_number avalue prec 3
    @    s$set_number avalue prepend ""
    @    s$set_number avalue append ""
    @    s$set_number avalue offset 0.000000 , 0.000000
    @    s$set_number errorbar on
    @    s$set_number errorbar place both
    @    s$set_number errorbar color $color_number
    @    s$set_number errorbar pattern 1
    @    s$set_number errorbar size 1.000000
    @    s$set_number errorbar linewidth 1.0
    @    s$set_number errorbar linestyle 1
    @    s$set_number errorbar riser linewidth 1.0
    @    s$set_number errorbar riser linestyle 1
    @    s$set_number errorbar riser clip off
    @    s$set_number errorbar riser clip length 0.100000
    @    s$set_number comment "Cols 1:2"
    @    s$set_number legend "$legend"
    """
)

AGR_SINGLESET_TEMPLATE = Template("""
    @target G0.S$set_number
    @type xy
    $xydata
    """)

MATPLOTLIB_HEADER_AGG_TEMPLATE = Template(
    """# -*- coding: utf-8 -*-

import matplotlib
matplotlib.use('Agg')

from matplotlib import rc
# Uncomment to change default font
#rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
rc('font', **{'family': 'serif', 'serif': ['Computer Modern', 'CMU Serif', 'Times New Roman', 'DejaVu Serif']})
# To use proper font for, e.g., Gamma if usetex is set to False
rc('mathtext', fontset='cm')

rc('text', usetex=True)

import pylab as pl

# I use json to make sure the input is sanitized
import json

print_comment = False
"""
)

MATPLOTLIB_HEADER_TEMPLATE = Template(
    """# -*- coding: utf-8 -*-

from matplotlib import rc
# Uncomment to change default font
#rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
rc('font', **{'family': 'serif', 'serif': ['Computer Modern', 'CMU Serif', 'Times New Roman', 'DejaVu Serif']})
# To use proper font for, e.g., Gamma if usetex is set to False
rc('mathtext', fontset='cm')

rc('text', usetex=True)

import pylab as pl

# I use json to make sure the input is sanitized
import json

print_comment = False
"""
)

MATPLOTLIB_IMPORT_DATA_INLINE_TEMPLATE = Template('''all_data_str = r"""$all_data_json"""
''')

MATPLOTLIB_IMPORT_DATA_FROMFILE_TEMPLATE = Template(
    """with open("$json_fname", encoding='utf8') as f:
    all_data_str = f.read()
"""
)

MULTI_KP = """
for path in paths:
    if path['length'] <= 1:
        # Avoid printing empty lines
        continue
    x = path['x']
    #for band in bands:
    for band, band_type in zip(path['values'], all_data['band_type_idx']):

        # For now we support only two colors
        if band_type % 2 == 0:
            further_plot_options = further_plot_options1
        else:
            further_plot_options = further_plot_options2

        # Put the legend text only once
        label = None
        if first_band_1 and band_type % 2 == 0:
            first_band_1 = False
            label = all_data.get('legend_text', None)
        elif first_band_2 and band_type % 2 == 1:
            first_band_2 = False
            label = all_data.get('legend_text2', None)

        p.plot(x, band, label=label,
               **further_plot_options
        )
"""

SINGLE_KP = """
path = paths[0]
values = path['values']
x = [path['x'] for _ in values]
p.scatter(x, values, marker="_")
"""

MATPLOTLIB_BODY_TEMPLATE = Template(
    """all_data = json.loads(all_data_str)

if not all_data.get('use_latex', False):
    rc('text', usetex=False)

#x = all_data['x']
#bands = all_data['bands']
paths = all_data['paths']
tick_pos = all_data['tick_pos']
tick_labels = all_data['tick_labels']

# Option for bands (all, or those of type 1 if there are two spins)
further_plot_options1 = {}
further_plot_options1['color'] = all_data.get('bands_color', 'k')
further_plot_options1['linewidth'] = all_data.get('bands_linewidth', 0.5)
further_plot_options1['linestyle'] = all_data.get('bands_linestyle', None)
further_plot_options1['marker'] = all_data.get('bands_marker', None)
further_plot_options1['markersize'] = all_data.get('bands_markersize', None)
further_plot_options1['markeredgecolor'] = all_data.get('bands_markeredgecolor', None)
further_plot_options1['markeredgewidth'] = all_data.get('bands_markeredgewidth', None)
further_plot_options1['markerfacecolor'] = all_data.get('bands_markerfacecolor', None)

# Options for second-type of bands if present (e.g. spin up vs. spin down)
further_plot_options2 = {}
further_plot_options2['color'] = all_data.get('bands_color2', 'r')
# Use the values of further_plot_options1 by default
further_plot_options2['linewidth'] = all_data.get('bands_linewidth2',
    further_plot_options1['linewidth']
    )
further_plot_options2['linestyle'] = all_data.get('bands_linestyle2',
    further_plot_options1['linestyle']
    )
further_plot_options2['marker'] = all_data.get('bands_marker2',
    further_plot_options1['marker']
    )
further_plot_options2['markersize'] = all_data.get('bands_markersize2',
    further_plot_options1['markersize']
    )
further_plot_options2['markeredgecolor'] = all_data.get('bands_markeredgecolor2',
    further_plot_options1['markeredgecolor']
    )
further_plot_options2['markeredgewidth'] = all_data.get('bands_markeredgewidth2',
    further_plot_options1['markeredgewidth']
    )
further_plot_options2['markerfacecolor'] = all_data.get('bands_markerfacecolor2',
    further_plot_options1['markerfacecolor']
    )

fig = pl.figure()
p = fig.add_subplot(1,1,1)

first_band_1 = True
first_band_2 = True

${plot_code}

p.set_xticks(tick_pos)
p.set_xticklabels(tick_labels)
p.set_xlim([all_data['x_min_lim'], all_data['x_max_lim']])
p.set_ylim([all_data['y_min_lim'], all_data['y_max_lim']])
p.xaxis.grid(True, which='major', color='#888888', linestyle='-', linewidth=0.5)

if all_data.get('plot_zero_axis', False):
    p.axhline(
        0.,
        color=all_data.get('zero_axis_color', '#888888'),
        linestyle=all_data.get('zero_axis_linestyle', '--'),
        linewidth=all_data.get('zero_axis_linewidth', 0.5),
        )
if all_data['title']:
    p.set_title(all_data['title'])
if all_data['legend_text']:
    p.legend(loc='best')
p.set_ylabel(all_data['yaxis_label'])

try:
    if print_comment:
        print(all_data['comment'])
except KeyError:
    pass
"""
)

MATPLOTLIB_FOOTER_TEMPLATE_SHOW = Template("""pl.show()""")

MATPLOTLIB_FOOTER_TEMPLATE_EXPORTFILE = Template("""pl.savefig("$fname", format="$format")""")

MATPLOTLIB_FOOTER_TEMPLATE_EXPORTFILE_WITH_DPI = Template("""pl.savefig("$fname", format="$format", dpi=$dpi)""")


def get_bands_and_parents_structure(args, backend=None):
    """Search for bands and return bands and the closest structure that is a parent of the instance.

    :returns:
        A list of sublists, each latter containing (in order):
            pk as string, formula as string, creation date, bandsdata-label
    """
    # pylint: disable=too-many-locals,too-many-branches

    import datetime

    from aiida import orm
    from aiida.common import timezone

    q_build = orm.QueryBuilder(backend=backend)
    if args.all_users is False:
        q_build.append(orm.User, tag='creator', filters={'email': orm.User.collection.get_default().email})
    else:
        q_build.append(orm.User, tag='creator')

    group_filters = {}
    with_args = {}

    if args.group_name is not None:
        group_filters.update({'label': {'in': args.group_name}})
    if args.group_pk is not None:
        group_filters.update({'id': {'in': args.group_pk}})

    if group_filters:
        q_build.append(orm.Group, tag='group', filters=group_filters, with_user='creator')
        with_args = {'with_group': 'group'}
    else:
        # Note: This is a workaround for the QB constraint of not allowing multiple ``with_*`` criteria. Correctly we
        # would like to specify with_user always on the ``BandsData`` directly and optionally add with_group. Until this
        # is resolved, add the ``with_user`` on the group if specified and on the ``BandsData`` if not.
        with_args = {'with_user': 'creator'}

    bdata_filters = {}
    if args.past_days is not None:
        bdata_filters.update({'ctime': {'>=': timezone.now() - datetime.timedelta(days=args.past_days)}})

    q_build.append(orm.BandsData, tag='bdata', filters=bdata_filters, project=['id', 'label', 'ctime'], **with_args)
    bands_list_data = q_build.all()

    q_build.append(
        orm.StructureData,
        tag='sdata',
        with_descendants='bdata',
        # We don't care about the creator of StructureData
        project=['id', 'attributes.kinds', 'attributes.sites']
    )

    q_build.order_by({orm.StructureData: {'ctime': 'desc'}})

    structure_dict = {}
    list_data = q_build.distinct().all()
    for bid, _, _, _, akinds, asites in list_data:
        structure_dict[bid] = (akinds, asites)

    entry_list = []
    already_visited_bdata = set()

    for [bid, blabel, bdate] in bands_list_data:

        # We process only one StructureData per BandsData.
        # We want to process the closest StructureData to
        # every BandsData.
        # We hope that the StructureData with the latest
        # creation time is the closest one.
        # This will be updated when the QueryBuilder supports
        # order_by by the distance of two nodes.
        if already_visited_bdata.__contains__(bid):
            continue
        already_visited_bdata.add(bid)
        strct = structure_dict.get(bid, None)

        if strct is not None:
            akinds, asites = strct
            formula = _extract_formula(akinds, asites, args)
        else:
            if args.element is not None or args.element_only is not None:
                formula = None
            else:
                formula = '<<NOT FOUND>>'

        if formula is None:
            continue
        entry_list.append([str(bid), str(formula), bdate.strftime('%d %b %Y'), blabel])

    return entry_list


def _extract_formula(akinds, asites, args):
    """
    Extract formula from the structure object.

    :param akinds: list of kinds, e.g. [{'mass': 55.845, 'name': 'Fe', 'symbols': ['Fe'], 'weights': [1.0]},
                                        {'mass': 15.9994, 'name': 'O', 'symbols': ['O'], 'weights': [1.0]}]
    :param asites: list of structure sites e.g. [{'position': [0.0, 0.0, 0.0], 'kind_name': 'Fe'},
                                                    {'position': [2.0, 2.0, 2.0], 'kind_name': 'O'}]
    :param args: a namespace with parsed command line parameters, here only 'element' and 'element_only' are used
    :type args: dict

    :return: a string with formula if the formula is found
    """
    from aiida.orm.nodes.data.structure import get_formula, get_symbols_string

    if args.element is not None:
        all_symbols = [_['symbols'][0] for _ in akinds]
        if not any(s in args.element for s in all_symbols):
            return None

    if args.element_only is not None:
        all_symbols = [_['symbols'][0] for _ in akinds]
        if not all(s in all_symbols for s in args.element_only):
            return None

    # We want only the StructureData that have attributes
    if akinds is None or asites is None:
        return '<<UNKNOWN>>'

    symbol_dict = {}
    for k in akinds:
        symbols = k['symbols']
        weights = k['weights']
        symbol_dict[k['name']] = get_symbols_string(symbols, weights)

    try:
        symbol_list = []
        for site in asites:
            symbol_list.append(symbol_dict[site['kind_name']])
        formula = get_formula(symbol_list, mode=args.formulamode)
    # If for some reason there is no kind with the name
    # referenced by the site
    except KeyError:
        formula = '<<UNKNOWN>>'
    return formula



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utility functions to draw ASCII diagrams to the command line."""
__all__ = ('format_call_graph',)

TREE_LAST_ENTRY = '\u2514\u2500\u2500 '
TREE_MIDDLE_ENTRY = '\u251C\u2500\u2500 '
TREE_FIRST_ENTRY = TREE_MIDDLE_ENTRY


def calc_info(node):
    """Return a string with the summary of the state of a CalculationNode."""
    from aiida.orm import ProcessNode, WorkChainNode

    if not isinstance(node, ProcessNode):
        raise TypeError(f'Unknown type: {type(node)}')

    process_label = node.process_label
    process_state = node.process_state.value.capitalize()
    exit_status = node.exit_status

    if exit_status is not None:
        string = f'{process_label}<{node.pk}> {process_state} [{exit_status}]'
    else:
        string = f'{process_label}<{node.pk}> {process_state}'

    if isinstance(node, WorkChainNode) and node.stepper_state_info:
        string += f' [{node.stepper_state_info}]'

    return string


def format_call_graph(calc_node, info_fn=calc_info):
    """
    Print a tree like the POSIX tree command for the calculation call graph

    :param calc_node: The calculation node
    :param info_fn: An optional function that takes the node and returns a string
        of information to be displayed for each node.
    """
    call_tree = build_call_graph(calc_node, info_fn=info_fn)
    return format_tree_descending(call_tree)


def build_call_graph(calc_node, info_fn=calc_info):
    """Build the call graph of a given node."""
    info_string = info_fn(calc_node)
    called = calc_node.called
    called.sort(key=lambda x: x.ctime)
    if called:
        return info_string, [build_call_graph(child, info_fn) for child in called]

    return info_string


def format_tree_descending(tree, prefix='', pos=-1):
    """Format a descending tree."""
    # pylint: disable=too-many-branches
    text = []

    if isinstance(tree, tuple):
        info = tree[0]
    else:
        info = tree

    if pos == -1:
        pre = ''
    elif pos == 0:
        pre = f'{prefix}{TREE_FIRST_ENTRY}'
    elif pos == 1:
        pre = f'{prefix}{TREE_MIDDLE_ENTRY}'
    else:
        pre = f'{prefix}{TREE_LAST_ENTRY}'
    text.append(f'{pre}{info}')

    if isinstance(tree, tuple):
        _, value = tree
        num_entries = len(value)
        if pos in [-1, 2]:
            new_prefix = f'{prefix}    '
        else:
            new_prefix = f'{prefix}│   '
        for i, entry in enumerate(value):
            if i == num_entries - 1:
                pos = 2
            elif i == 0:
                pos = 0
            else:
                pos = 1
            text.append(format_tree_descending(entry, new_prefix, pos))

    return '\n'.join(text)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Add indexes to dbworkflowdata table

Revision ID: 89176227b25
Revises:
Create Date: 2017-11-03 11:06:00.327195

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = '89176227b25'
down_revision = 'a6048f0ffca8'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    op.create_index('ix_db_dbworkflowdata_aiida_obj_id', 'db_dbworkflowdata', ['aiida_obj_id'])
    op.create_index('ix_db_dbworkflowdata_parent_id', 'db_dbworkflowdata', ['parent_id'])


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of 89176227b25.')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module containing utilities and classes relating to job calculations running on systems that require transport."""
import asyncio
import contextlib
import contextvars
import logging
import time
from typing import TYPE_CHECKING, Any, Dict, Hashable, Iterator, List, Optional

from aiida.common import lang
from aiida.orm import AuthInfo

if TYPE_CHECKING:
    from aiida.engine.transports import TransportQueue
    from aiida.schedulers.datastructures import JobInfo

__all__ = ('JobsList', 'JobManager')


class JobsList:
    """Manager of calculation jobs submitted with a specific ``AuthInfo``, i.e. computer configured for a specific user.

    This container of active calculation jobs is used to update their status periodically in batches, ensuring that
    even when a lot of jobs are running, the scheduler update command is not triggered for each job individually.

    In addition, the :py:class:`~aiida.orm.computers.Computer` for which the :py:class:`~aiida.orm.authinfos.AuthInfo`
    is configured, can define a minimum polling interval. This class will guarantee that the time between update calls
    to the scheduler is larger or equal to that minimum interval.

    Note that since each instance operates on a specific authinfo, the guarantees of batching scheduler update calls
    and the limiting of number of calls per unit time, through the minimum polling interval, is only applicable for jobs
    launched with that particular authinfo. If multiple authinfo instances with the same computer, have active jobs
    these limitations are not respected between them, since there is no communication between ``JobsList`` instances.
    See the :py:class:`~aiida.engine.processes.calcjobs.manager.JobManager` for example usage.
    """

    def __init__(self, authinfo: AuthInfo, transport_queue: 'TransportQueue', last_updated: Optional[float] = None):
        """Construct an instance for the given authinfo and transport queue.

        :param authinfo: The authinfo used to check the jobs list
        :param transport_queue: A transport queue
        :param last_updated: initialize the last updated timestamp

        """
        lang.type_check(last_updated, float, allow_none=True)

        self._authinfo = authinfo
        self._transport_queue = transport_queue
        self._loop = transport_queue.loop
        self._logger = logging.getLogger(__name__)

        self._jobs_cache: Dict[Hashable, 'JobInfo'] = {}
        self._job_update_requests: Dict[Hashable, asyncio.Future] = {}  # Mapping: {job_id: Future}
        self._last_updated = last_updated
        self._update_handle: Optional[asyncio.TimerHandle] = None

    @property
    def logger(self) -> logging.Logger:
        """Return the logger configured for this instance.

        :return: the logger
        """
        return self._logger

    def get_minimum_update_interval(self) -> float:
        """Get the minimum interval that should be respected between updates of the list.

        :return: the minimum interval

        """
        return self._authinfo.computer.get_minimum_job_poll_interval()

    @property
    def last_updated(self) -> Optional[float]:
        """Get the timestamp of when the list was last updated as produced by `time.time()`

        :return: The last update point

        """
        return self._last_updated

    async def _get_jobs_from_scheduler(self) -> Dict[Hashable, 'JobInfo']:
        """Get the current jobs list from the scheduler.

        :return: a mapping of job ids to :py:class:`~aiida.schedulers.datastructures.JobInfo` instances

        """
        with self._transport_queue.request_transport(self._authinfo) as request:
            self.logger.info('waiting for transport')
            transport = await request

            scheduler = self._authinfo.computer.get_scheduler()
            scheduler.set_transport(transport)

            kwargs: Dict[str, Any] = {'as_dict': True}
            if scheduler.get_feature('can_query_by_user'):
                kwargs['user'] = '$USER'
            else:
                kwargs['jobs'] = self._get_jobs_with_scheduler()

            scheduler_response = scheduler.get_jobs(**kwargs)

            # Update the last update time and clear the jobs cache
            self._last_updated = time.time()
            jobs_cache = {}
            self.logger.info(f'AuthInfo<{self._authinfo.pk}>: successfully retrieved status of active jobs')

            for job_id, job_info in scheduler_response.items():
                jobs_cache[job_id] = job_info

            return jobs_cache

    async def _update_job_info(self) -> None:
        """Update all of the job information objects.

        This will set the futures for all pending update requests where the corresponding job has a new status compared
        to the last update.
        """
        try:
            if not self._update_requests_outstanding():
                return

            # Update our cache of the job states
            self._jobs_cache = await self._get_jobs_from_scheduler()
        except Exception as exception:
            # Set the exception on all the update futures
            for future in self._job_update_requests.values():
                if not future.done():
                    future.set_exception(exception)

            # Reset the `_update_handle` manually. Normally this is done in the `updating` coroutine, but since we
            # reraise this exception, that code path is never hit. If the next time a request comes in, the method
            # `_ensure_updating` will falsely conclude we are still updating, since the handle is not `None` and so it
            # will not schedule the next update, causing the job update futures to never be resolved.
            self._update_handle = None

            raise
        else:
            for job_id, future in self._job_update_requests.items():
                if not future.done():
                    future.set_result(self._jobs_cache.get(job_id, None))
        finally:
            self._job_update_requests = {}

    @contextlib.contextmanager
    def request_job_info_update(self, job_id: Hashable) -> Iterator['asyncio.Future[JobInfo]']:
        """Request job info about a job when the job next changes state.

        If the job is not found in the jobs list at the update, the future will resolve to `None`.

        :param job_id: job identifier
        :return: future that will resolve to a `JobInfo` object when the job changes state
        """
        # Get or create the future
        request = self._job_update_requests.setdefault(job_id, asyncio.Future())
        assert not request.done(), 'Expected pending job info future, found in done state.'

        try:
            self._ensure_updating()
            yield request
        finally:
            pass

    def _ensure_updating(self) -> None:
        """Ensure that we are updating the job list from the remote resource.

        This will automatically stop if there are no outstanding requests.
        """

        async def updating():
            """Do the actual update, stop if not requests left."""
            await self._update_job_info()
            # Any outstanding requests?
            if self._update_requests_outstanding():
                self._update_handle = self._loop.call_later(
                    self._get_next_update_delay(),
                    asyncio.ensure_future,
                    updating(),
                    context=contextvars.Context(),  #  type: ignore[call-arg]
                )
            else:
                self._update_handle = None

        # Check if we're already updating
        if self._update_handle is None:
            self._update_handle = self._loop.call_later(
                self._get_next_update_delay(),
                asyncio.ensure_future,
                updating(),
                context=contextvars.Context(),  #  type: ignore[call-arg]
            )

    @staticmethod
    def _has_job_state_changed(old: Optional['JobInfo'], new: Optional['JobInfo']) -> bool:
        """Return whether the states `old` and `new` are different.


        """
        if old is None and new is None:
            return False

        if old is None or new is None:
            # One is None and the other isn't
            return True

        return old.job_state != new.job_state or old.job_substate != new.job_substate

    def _get_next_update_delay(self) -> float:
        """Calculate when we are next allowed to poll the scheduler.

        This delay is calculated as the minimum polling interval defined by the authentication info for this instance,
        minus time elapsed since the last update.

        :return: delay (in seconds) after which the scheduler may be polled again

        """
        if self.last_updated is None:
            # Never updated, so do it straight away
            return 0.

        # Make sure to actually 'get' the minimum interval here, in case the user changed since last time
        minimum_interval = self.get_minimum_update_interval()
        elapsed = time.time() - self.last_updated

        delay = max(minimum_interval - elapsed, 0.)

        return delay

    def _update_requests_outstanding(self) -> bool:
        return any(not request.done() for request in self._job_update_requests.values())

    def _get_jobs_with_scheduler(self) -> List[str]:
        """Get all the jobs that are currently with scheduler.

        :return: the list of jobs with the scheduler
        :rtype: list
        """
        return [str(job_id) for job_id, _ in self._job_update_requests.items()]


class JobManager:
    """A manager for :py:class:`~aiida.engine.processes.calcjobs.calcjob.CalcJob` submitted to ``Computer`` instances.

    When a calculation job is submitted to a :py:class:`~aiida.orm.computers.Computer`, it actually uses a specific
    :py:class:`~aiida.orm.authinfos.AuthInfo`, which is a computer configured for a :py:class:`~aiida.orm.users.User`.
    The ``JobManager`` maintains a mapping of :py:class:`~aiida.engine.processes.calcjobs.manager.JobsList` instances
    for each authinfo that has active calculation jobs. These jobslist instances are then responsible for bundling
    scheduler updates for all the jobs they maintain (i.e. that all share the same authinfo) and update their status.

    As long as a :py:class:`~aiida.engine.runners.Runner` will create a single ``JobManager`` instance and use that for
    its lifetime, the guarantees made by the ``JobsList`` about respecting the minimum polling interval of the scheduler
    will be maintained. Note, however, that since each ``Runner`` will create its own job manager, these guarantees
    only hold per runner.
    """

    def __init__(self, transport_queue: 'TransportQueue') -> None:
        self._transport_queue = transport_queue
        self._job_lists: Dict[Hashable, 'JobInfo'] = {}

    def get_jobs_list(self, authinfo: AuthInfo) -> JobsList:
        """Get or create a new `JobLists` instance for the given authinfo.

        :param authinfo: the `AuthInfo`
        :return: a `JobsList` instance
        """
        if authinfo.pk not in self._job_lists:
            self._job_lists[authinfo.pk] = JobsList(authinfo, self._transport_queue)

        return self._job_lists[authinfo.pk]

    @contextlib.contextmanager
    def request_job_info_update(self, authinfo: AuthInfo, job_id: Hashable) -> Iterator['asyncio.Future[JobInfo]']:
        """Get a future that will resolve to information about a given job.

        This is a context manager so that if the user leaves the context the request is automatically cancelled.

        """
        with self.get_jobs_list(authinfo).request_job_info_update(job_id) as request:
            try:
                yield request
            finally:
                if not request.done():
                    request.cancel()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module with `Node` sub class for calculation job processes."""
import datetime
from typing import TYPE_CHECKING, Any, AnyStr, Dict, List, Optional, Sequence, Tuple, Type, Union

from aiida.common import exceptions
from aiida.common.datastructures import CalcJobState
from aiida.common.lang import classproperty
from aiida.common.links import LinkType

from ..process import ProcessNodeCaching
from .calculation import CalculationNode

if TYPE_CHECKING:
    from aiida.engine.processes.builder import ProcessBuilder
    from aiida.orm import FolderData
    from aiida.orm.authinfos import AuthInfo
    from aiida.orm.utils.calcjob import CalcJobResultManager
    from aiida.parsers import Parser
    from aiida.schedulers.datastructures import JobInfo, JobState
    from aiida.tools.calculations import CalculationTools
    from aiida.transports import Transport

__all__ = ('CalcJobNode',)


class CalcJobNodeCaching(ProcessNodeCaching):
    """Interface to control caching of a node instance."""

    def _get_objects_to_hash(self) -> List[Any]:
        """Return a list of objects which should be included in the hash.

        This method is purposefully overridden from the base `Node` class, because we do not want to include the
        repository folder in the hash. The reason is that the hash of this node is computed in the `store` method, at
        which point the input files that will be stored in the repository have not yet been generated. Including these
        anyway in the computation of the hash would mean that the hash of the node would change as soon as the process
        has started and the input files have been written to the repository.
        """
        from importlib import import_module
        objects = [
            import_module(self._node.__module__.split('.', 1)[0]).__version__,
            {
                key: val
                for key, val in self._node.base.attributes.items()
                if key not in self._node._hash_ignored_attributes and key not in self._node._updatable_attributes  # pylint: disable=unsupported-membership-test,protected-access
            },
            self._node.computer.uuid if self._node.computer is not None else None,  # pylint: disable=no-member
            {
                entry.link_label: entry.node.base.caching.get_hash()
                for entry in self._node.base.links.get_incoming(link_type=(LinkType.INPUT_CALC, LinkType.INPUT_WORK))
                if entry.link_label not in self._hash_ignored_inputs
            }
        ]
        return objects


class CalcJobNode(CalculationNode):
    """ORM class for all nodes representing the execution of a CalcJob."""

    # pylint: disable=too-many-public-methods

    CALC_JOB_STATE_KEY = 'state'
    IMMIGRATED_KEY = 'imported'
    REMOTE_WORKDIR_KEY = 'remote_workdir'
    RETRIEVE_LIST_KEY = 'retrieve_list'
    RETRIEVE_TEMPORARY_LIST_KEY = 'retrieve_temporary_list'
    SCHEDULER_JOB_ID_KEY = 'job_id'
    SCHEDULER_STATE_KEY = 'scheduler_state'
    SCHEDULER_LAST_CHECK_TIME_KEY = 'scheduler_lastchecktime'
    SCHEDULER_LAST_JOB_INFO_KEY = 'last_job_info'
    SCHEDULER_DETAILED_JOB_INFO_KEY = 'detailed_job_info'

    # An optional entry point for a CalculationTools instance
    _tools = None

    @property
    def tools(self) -> 'CalculationTools':
        """Return the calculation tools that are registered for the process type associated with this calculation.

        If the entry point name stored in the `process_type` of the CalcJobNode has an accompanying entry point in the
        `aiida.tools.calculations` entry point category, it will attempt to load the entry point and instantiate it
        passing the node to the constructor. If the entry point does not exist, cannot be resolved or loaded, a warning
        will be logged and the base CalculationTools class will be instantiated and returned.

        :return: CalculationTools instance
        """
        from aiida.plugins.entry_point import get_entry_point_from_string, is_valid_entry_point_string, load_entry_point
        from aiida.tools.calculations import CalculationTools

        if self._tools is None:
            entry_point_string = self.process_type

            if entry_point_string and is_valid_entry_point_string(entry_point_string):
                entry_point = get_entry_point_from_string(entry_point_string)

                try:
                    tools_class = load_entry_point(
                        'aiida.tools.calculations',
                        entry_point.name  # type: ignore[attr-defined]
                    )
                    self._tools = tools_class(self)
                except exceptions.EntryPointError as exception:
                    self._tools = CalculationTools(self)
                    entry_point_name = entry_point.name  # type: ignore[attr-defined]
                    self.logger.warning(
                        f'could not load the calculation tools entry point {entry_point_name}: {exception}'
                    )

        return self._tools

    @classproperty
    def _updatable_attributes(cls) -> Tuple[str, ...]:  # pylint: disable=no-self-argument
        return super()._updatable_attributes + (
            cls.CALC_JOB_STATE_KEY,
            cls.IMMIGRATED_KEY,
            cls.REMOTE_WORKDIR_KEY,
            cls.RETRIEVE_LIST_KEY,
            cls.RETRIEVE_TEMPORARY_LIST_KEY,
            cls.SCHEDULER_JOB_ID_KEY,
            cls.SCHEDULER_STATE_KEY,
            cls.SCHEDULER_LAST_CHECK_TIME_KEY,
            cls.SCHEDULER_LAST_JOB_INFO_KEY,
            cls.SCHEDULER_DETAILED_JOB_INFO_KEY,
        )

    @classproperty
    def _hash_ignored_attributes(cls) -> Tuple[str, ...]:  # pylint: disable=no-self-argument
        return super()._hash_ignored_attributes + (
            'queue_name',
            'account',
            'qos',
            'priority',
            'max_wallclock_seconds',
            'max_memory_kb',
        )

    def get_builder_restart(self) -> 'ProcessBuilder':
        """Return a `ProcessBuilder` that is ready to relaunch the same `CalcJob` that created this node.

        The process class will be set based on the `process_type` of this node and the inputs of the builder will be
        prepopulated with the inputs registered for this node. This functionality is very useful if a process has
        completed and you want to relaunch it with slightly different inputs.

        In addition to prepopulating the input nodes, which is implemented by the base `ProcessNode` class, here we
        also add the `options` that were passed in the `metadata` input of the `CalcJob` process.

        """
        builder = super().get_builder_restart()
        builder.metadata.options = self.get_options()  # type: ignore[attr-defined]
        return builder

    @property
    def is_imported(self) -> bool:
        """Return whether the calculation job was imported instead of being an actual run."""
        return self.base.attributes.get(self.IMMIGRATED_KEY, None) is True

    def get_option(self, name: str) -> Optional[Any]:
        """
        Retun the value of an option that was set for this CalcJobNode

        :param name: the option name
        :return: the option value or None
        :raises: ValueError for unknown option
        """
        return self.base.attributes.get(name, None)

    def set_option(self, name: str, value: Any) -> None:
        """
        Set an option to the given value

        :param name: the option name
        :param value: the value to set
        :raises: ValueError for unknown option
        :raises: TypeError for values with invalid type
        """
        self.base.attributes.set(name, value)

    def get_options(self) -> Dict[str, Any]:
        """
        Return the dictionary of options set for this CalcJobNode

        :return: dictionary of the options and their values
        """
        options = {}
        for name in self.process_class.spec_options.keys():  # type: ignore[attr-defined]
            value = self.get_option(name)
            if value is not None:
                options[name] = value

        return options

    def set_options(self, options: Dict[str, Any]) -> None:
        """
        Set the options for this CalcJobNode

        :param options: dictionary of option and their values to set
        """
        for name, value in options.items():
            self.set_option(name, value)

    def get_state(self) -> Optional[CalcJobState]:
        """Return the calculation job active sub state.

        The calculation job state serves to give more granular state information to `CalcJobs`, in addition to the
        generic process state, while the calculation job is active. The state can take values from the enumeration
        defined in `aiida.common.datastructures.CalcJobState` and can be used to query for calculation jobs in specific
        active states.

        :return: instance of `aiida.common.datastructures.CalcJobState` or `None` if invalid value, or not set
        """
        state = self.base.attributes.get(self.CALC_JOB_STATE_KEY, None)

        try:
            state = CalcJobState(state)
        except ValueError:
            state = None

        return state

    def set_state(self, state: CalcJobState) -> None:
        """Set the calculation active job state.

        :raise: ValueError if state is invalid
        """
        if not isinstance(state, CalcJobState):
            raise ValueError(f'{state} is not a valid CalcJobState')

        self.base.attributes.set(self.CALC_JOB_STATE_KEY, state.value)

    def delete_state(self) -> None:
        """Delete the calculation job state attribute if it exists."""
        try:
            self.base.attributes.delete(self.CALC_JOB_STATE_KEY)
        except AttributeError:
            pass

    def set_remote_workdir(self, remote_workdir: str) -> None:
        """Set the absolute path to the working directory on the remote computer where the calculation is run.

        :param remote_workdir: absolute filepath to the remote working directory
        """
        self.base.attributes.set(self.REMOTE_WORKDIR_KEY, remote_workdir)

    def get_remote_workdir(self) -> Optional[str]:
        """Return the path to the remote (on cluster) scratch folder of the calculation.

        :return: a string with the remote path
        """
        return self.base.attributes.get(self.REMOTE_WORKDIR_KEY, None)

    @staticmethod
    def _validate_retrieval_directive(directives: Sequence[Union[str, Tuple[str, str, str]]]) -> None:
        """Validate a list or tuple of file retrieval directives.

        :param directives: a list or tuple of file retrieval directives
        :raise ValueError: if the format of the directives is invalid
        """
        if not isinstance(directives, (tuple, list)):
            raise TypeError('file retrieval directives has to be a list or tuple')

        for directive in directives:

            # A string as a directive is valid, so we continue
            if isinstance(directive, str):
                continue

            # Otherwise, it has to be a tuple of length three with specific requirements
            if not isinstance(directive, (tuple, list)) or len(directive) != 3:
                raise ValueError(f'invalid directive, not a list or tuple of length three: {directive}')

            if not isinstance(directive[0], str):
                raise ValueError('invalid directive, first element has to be a string representing remote path')

            if not isinstance(directive[1], str):
                raise ValueError('invalid directive, second element has to be a string representing local path')

            if not isinstance(directive[2], int):
                raise ValueError('invalid directive, three element has to be an integer representing the depth')

    def set_retrieve_list(self, retrieve_list: Sequence[Union[str, Tuple[str, str, str]]]) -> None:
        """Set the retrieve list.

        This list of directives will instruct the daemon what files to retrieve after the calculation has completed.
        list or tuple of files or paths that should be retrieved by the daemon.

        :param retrieve_list: list or tuple of with filepath directives
        """
        self._validate_retrieval_directive(retrieve_list)
        self.base.attributes.set(self.RETRIEVE_LIST_KEY, retrieve_list)

    def get_retrieve_list(self) -> Optional[Sequence[Union[str, Tuple[str, str, str]]]]:
        """Return the list of files/directories to be retrieved on the cluster after the calculation has completed.

        :return: a list of file directives
        """
        return self.base.attributes.get(self.RETRIEVE_LIST_KEY, None)

    def set_retrieve_temporary_list(self, retrieve_temporary_list: Sequence[Union[str, Tuple[str, str, str]]]) -> None:
        """Set the retrieve temporary list.

        The retrieve temporary list stores files that are retrieved after completion and made available during parsing
        and are deleted as soon as the parsing has been completed.

        :param retrieve_temporary_list: list or tuple of with filepath directives
        """
        self._validate_retrieval_directive(retrieve_temporary_list)
        self.base.attributes.set(self.RETRIEVE_TEMPORARY_LIST_KEY, retrieve_temporary_list)

    def get_retrieve_temporary_list(self) -> Optional[Sequence[Union[str, Tuple[str, str, str]]]]:
        """Return list of files to be retrieved from the cluster which will be available during parsing.

        :return: a list of file directives
        """
        return self.base.attributes.get(self.RETRIEVE_TEMPORARY_LIST_KEY, None)

    def set_job_id(self, job_id: Union[int, str]) -> None:
        """Set the job id that was assigned to the calculation by the scheduler.

        .. note:: the id will always be stored as a string

        :param job_id: the id assigned by the scheduler after submission
        """
        return self.base.attributes.set(self.SCHEDULER_JOB_ID_KEY, str(job_id))

    def get_job_id(self) -> Optional[str]:
        """Return job id that was assigned to the calculation by the scheduler.

        :return: the string representation of the scheduler job id
        """
        return self.base.attributes.get(self.SCHEDULER_JOB_ID_KEY, None)

    def set_scheduler_state(self, state: 'JobState') -> None:
        """Set the scheduler state.

        :param state: an instance of `JobState`
        """
        from aiida.common import timezone
        from aiida.schedulers.datastructures import JobState

        if not isinstance(state, JobState):
            raise ValueError(f'scheduler state should be an instance of JobState, got: {state}')

        self.base.attributes.set(self.SCHEDULER_STATE_KEY, state.value)
        self.base.attributes.set(self.SCHEDULER_LAST_CHECK_TIME_KEY, timezone.now().isoformat())

    def get_scheduler_state(self) -> Optional['JobState']:
        """Return the status of the calculation according to the cluster scheduler.

        :return: a JobState enum instance.
        """
        from aiida.schedulers.datastructures import JobState

        state = self.base.attributes.get(self.SCHEDULER_STATE_KEY, None)

        if state is None:
            return state

        return JobState(state)

    def get_scheduler_lastchecktime(self) -> Optional[datetime.datetime]:
        """Return the time of the last update of the scheduler state by the daemon or None if it was never set.

        :return: a datetime object or None
        """
        value = self.base.attributes.get(self.SCHEDULER_LAST_CHECK_TIME_KEY, None)

        if value is not None:
            value = datetime.datetime.fromisoformat(value)

        return value

    def set_detailed_job_info(self, detailed_job_info: Optional[dict]) -> None:
        """Set the detailed job info dictionary.

        :param detailed_job_info: a dictionary with metadata with the accounting of a completed job
        """
        self.base.attributes.set(self.SCHEDULER_DETAILED_JOB_INFO_KEY, detailed_job_info)

    def get_detailed_job_info(self) -> Optional[dict]:
        """Return the detailed job info dictionary.

        The scheduler is polled for the detailed job info after the job is completed and ready to be retrieved.

        :return: the dictionary with detailed job info if defined or None
        """
        return self.base.attributes.get(self.SCHEDULER_DETAILED_JOB_INFO_KEY, None)

    def set_last_job_info(self, last_job_info: 'JobInfo') -> None:
        """Set the last job info.

        :param last_job_info: a `JobInfo` object
        """
        self.base.attributes.set(self.SCHEDULER_LAST_JOB_INFO_KEY, last_job_info.get_dict())

    def get_last_job_info(self) -> Optional['JobInfo']:
        """Return the last information asked to the scheduler about the status of the job.

        The last job info is updated on every poll of the scheduler, except for the final poll when the job drops from
        the scheduler's job queue.
        For completed jobs, the last job info therefore contains the "second-to-last" job info that still shows the job
        as running. Please use :meth:`~aiida.orm.nodes.process.calculation.calcjob.CalcJobNode.get_detailed_job_info`
        instead.

        :return: a `JobInfo` object (that closely resembles a dictionary) or None.
        """
        from aiida.schedulers.datastructures import JobInfo

        last_job_info_dictserialized = self.base.attributes.get(self.SCHEDULER_LAST_JOB_INFO_KEY, None)

        if last_job_info_dictserialized is not None:
            job_info = JobInfo.load_from_dict(last_job_info_dictserialized)
        else:
            job_info = None

        return job_info

    def get_authinfo(self) -> 'AuthInfo':
        """Return the `AuthInfo` that is configured for the `Computer` set for this node.

        :return: `AuthInfo`
        """
        computer = self.computer

        if computer is None:
            raise exceptions.NotExistent('No computer has been set for this calculation')

        return computer.get_authinfo(self.user)  # pylint: disable=no-member

    def get_transport(self) -> 'Transport':
        """Return the transport for this calculation.

        :return: `Transport` configured with the `AuthInfo` associated to the computer of this node
        """
        return self.get_authinfo().get_transport()

    def get_parser_class(self) -> Optional[Type['Parser']]:
        """Return the output parser object for this calculation or None if no parser is set.

        :return: a `Parser` class.
        :raises `aiida.common.exceptions.EntryPointError`: if the parser entry point can not be resolved.
        """
        from aiida.plugins import ParserFactory

        parser_name = self.get_option('parser_name')

        if parser_name is not None:
            return ParserFactory(parser_name)

        return None

    @property
    def link_label_retrieved(self) -> str:
        """Return the link label used for the retrieved FolderData node."""
        return 'retrieved'

    def get_retrieved_node(self) -> Optional['FolderData']:
        """Return the retrieved data folder.

        :return: the retrieved FolderData node or None if not found
        """
        from aiida.orm import FolderData
        try:
            return self.base.links.get_outgoing(node_class=FolderData,
                                                link_label_filter=self.link_label_retrieved).one().node
        except ValueError:
            return None

    @property
    def res(self) -> 'CalcJobResultManager':
        """
        To be used to get direct access to the parsed parameters.

        :return: an instance of the CalcJobResultManager.

        :note: a practical example on how it is meant to be used: let's say that there is a key 'energy'
            in the dictionary of the parsed results which contains a list of floats.
            The command `calc.res.energy` will return such a list.
        """
        from aiida.orm.utils.calcjob import CalcJobResultManager
        return CalcJobResultManager(self)

    def get_scheduler_stdout(self) -> Optional[AnyStr]:
        """Return the scheduler stderr output if the calculation has finished and been retrieved, None otherwise.

        :return: scheduler stderr output or None
        """
        filename = self.get_option('scheduler_stdout')
        retrieved_node = self.get_retrieved_node()

        if filename is None or retrieved_node is None:
            return None

        try:
            stdout = retrieved_node.base.repository.get_object_content(filename)
        except IOError:
            stdout = None

        return stdout

    def get_scheduler_stderr(self) -> Optional[AnyStr]:
        """Return the scheduler stdout output if the calculation has finished and been retrieved, None otherwise.

        :return: scheduler stdout output or None
        """
        filename = self.get_option('scheduler_stderr')
        retrieved_node = self.get_retrieved_node()

        if filename is None or retrieved_node is None:
            return None

        try:
            stderr = retrieved_node.base.repository.get_object_content(filename)
        except IOError:
            stderr = None

        return stderr

    def get_description(self) -> str:
        """Return a description of the node based on its properties."""
        state = self.get_state()
        if not state:
            return ''
        return state.value



# -*- coding: utf-8 -*-
"""Module for the implementation of a file repository."""
import contextlib
import pathlib
from typing import Any, BinaryIO, Dict, Iterable, Iterator, List, Optional, Tuple, Union

from aiida.common.hashing import make_hash
from aiida.common.lang import type_check

from .backend import AbstractRepositoryBackend, SandboxRepositoryBackend
from .common import File, FileType

__all__ = ('Repository',)

FilePath = Union[str, pathlib.PurePosixPath]


class Repository:
    """File repository.

    This class provides an interface to a backend file repository instance, but unlike the backend repository, this
    class keeps a reference of the virtual file hierarchy. This means that through this interface, a client can create
    files and directories with a file hierarchy, just as they would on a local file system, except it is completely
    virtual as the files are stored by the backend which can store them in a completely flat structure. This also means
    that the internal virtual hierarchy of a ``Repository`` instance does not necessarily represent all the files that
    are stored by repository backend. The repository exposes a mere subset of all the file objects stored in the
    backend. This is why object deletion is also implemented as a soft delete, by default, where the files are just
    removed from the internal virtual hierarchy, but not in the actual backend. This is because those objects can be
    referenced by other instances.
    """

    # pylint: disable=too-many-public-methods

    _file_cls = File

    def __init__(self, backend: AbstractRepositoryBackend = None):
        """Construct a new instance with empty metadata.

        :param backend: instance of repository backend to use to actually store the file objects. By default, an
            instance of the ``SandboxRepositoryBackend`` will be created.
        """
        if backend is None:
            backend = SandboxRepositoryBackend()

        self.set_backend(backend)
        self.reset()

    def __str__(self) -> str:
        """Return the string representation of this repository."""
        return f'Repository<{str(self.backend)}>'

    @property
    def uuid(self) -> Optional[str]:
        """Return the unique identifier of the repository backend or ``None`` if it doesn't have one."""
        return self.backend.uuid

    @property
    def is_initialised(self) -> bool:
        """Return whether the repository backend has been initialised."""
        return self.backend.is_initialised

    @classmethod
    def from_serialized(cls, backend: AbstractRepositoryBackend, serialized: Dict[str, Any]) -> 'Repository':
        """Construct an instance where the metadata is initialized from the serialized content.

        :param backend: instance of repository backend to use to actually store the file objects.
        """
        instance = cls.__new__(cls)
        instance.__init__(backend)  # type: ignore[misc]

        if serialized:
            for name, obj in serialized['o'].items():
                instance.get_directory().objects[name] = cls._file_cls.from_serialized(obj, name)

        return instance

    def reset(self) -> None:
        self._directory = self._file_cls()

    def serialize(self) -> Dict[str, Any]:
        """Serialize the metadata into a JSON-serializable format.

        :return: dictionary with the content metadata.
        """
        return self._directory.serialize()

    @classmethod
    def flatten(cls, serialized=Optional[Dict[str, Any]], delimiter: str = '/') -> Dict[str, Optional[str]]:
        """Flatten the serialized content of a repository into a mapping of path -> key or None (if folder).

        Note, all folders are represented in the flattened output, and their path is suffixed with the delimiter.

        :param serialized: the serialized content of the repository.
        :param delimiter: the delimiter to use to separate the path elements.
        :return: dictionary with the flattened content.
        """
        if serialized is None:
            return {}
        items: Dict[str, Optional[str]] = {}
        stack = [('', serialized)]
        while stack:
            path, sub_dict = stack.pop()
            for name, obj in sub_dict.get('o', {}).items():
                sub_path = f'{path}{delimiter}{name}' if path else name
                if not obj:
                    items[f'{sub_path}{delimiter}'] = None
                elif 'k' in obj:
                    items[sub_path] = obj['k']
                else:
                    items[f'{sub_path}{delimiter}'] = None
                    stack.append((sub_path, obj))
        return items

    def hash(self) -> str:
        """Generate a hash of the repository's contents.

        .. warning:: this will read the content of all file objects contained within the virtual hierarchy into memory.

        :return: the hash representing the contents of the repository.
        """
        objects: Dict[str, Any] = {}
        for root, dirnames, filenames in self.walk():
            objects['__dirnames__'] = dirnames
            for filename in filenames:
                key = self.get_file(root / filename).key
                assert key is not None, 'Expected FileType.File to have a key'
                objects[str(root / filename)] = self.backend.get_object_hash(key)

        return make_hash(objects)

    @staticmethod
    def _pre_process_path(path: FilePath = None) -> pathlib.PurePosixPath:
        """Validate and convert the path to instance of ``pathlib.PurePosixPath``.

        This should be called by every method of this class before doing anything, such that it can safely assume that
        the path is a ``pathlib.PurePosixPath`` object, which makes path manipulation a lot easier.

        :param path: the path as a ``pathlib.PurePosixPath`` object or `None`.
        :raises TypeError: if the type of path was not a str nor a ``pathlib.PurePosixPath`` instance.
        """
        if path is None:
            return pathlib.PurePosixPath()

        if isinstance(path, str):
            path = pathlib.PurePosixPath(path)

        if not isinstance(path, pathlib.PurePosixPath):
            raise TypeError('path is not of type `str` nor `pathlib.PurePosixPath`.')

        if path.is_absolute():
            raise TypeError(f'path `{path}` is not a relative path.')

        return path

    @property
    def backend(self) -> AbstractRepositoryBackend:
        """Return the current repository backend.

        :return: the repository backend.
        """
        return self._backend

    def set_backend(self, backend: AbstractRepositoryBackend) -> None:
        """Set the backend for this repository.

        :param backend: the repository backend.
        :raises TypeError: if the type of the backend is invalid.
        """
        type_check(backend, AbstractRepositoryBackend)
        self._backend = backend

    def _insert_file(self, path: pathlib.PurePosixPath, key: str) -> None:
        """Insert a new file object in the object mapping.

        .. note:: this assumes the path is a valid relative path, so should be checked by the caller.

        :param path: the relative path where to store the object in the repository.
        :param key: fully qualified identifier for the object within the repository.
        """
        directory = self.create_directory(path.parent)
        directory.objects[path.name] = self._file_cls(path.name, FileType.FILE, key)

    def create_directory(self, path: FilePath) -> File:
        """Create a new directory with the given path.

        :param path: the relative path of the directory.
        :return: the created directory.
        :raises TypeError: if the path is not a string or ``Path``, or is an absolute path.
        """
        if path is None:
            raise TypeError('path cannot be `None`.')

        path = self._pre_process_path(path)
        directory = self._directory

        for part in path.parts:
            if part not in directory.objects:
                directory.objects[part] = self._file_cls(part)

            directory = directory.objects[part]

        return directory

    def get_file_keys(self) -> List[str]:
        """Return the keys of all file objects contained within this repository.

        :return: list of keys, which map a file to its content in the backend repository.
        """
        file_keys: List[str] = []

        def _add_file_keys(keys, objects):
            """Recursively add keys of all file objects to the keys list."""
            for obj in objects.values():
                if obj.file_type == FileType.FILE and obj.key is not None:
                    keys.append(obj.key)
                elif obj.file_type == FileType.DIRECTORY:
                    _add_file_keys(keys, obj.objects)

        _add_file_keys(file_keys, self._directory.objects)

        return file_keys

    def get_object(self, path: FilePath = None) -> File:
        """Return the object at the given path.

        :param path: the relative path where to store the object in the repository.
        :return: the `File` representing the object located at the given relative path.
        :raises TypeError: if the path is not a string or ``Path``, or is an absolute path.
        :raises FileNotFoundError: if no object exists for the given path.
        """
        path = self._pre_process_path(path)
        file_object = self._directory

        if not path.parts:
            return file_object

        for part in path.parts:
            if part not in file_object.objects:
                raise FileNotFoundError(f'object with path `{path}` does not exist.')

            file_object = file_object.objects[part]

        return file_object

    def get_directory(self, path: FilePath = None) -> File:
        """Return the directory object at the given path.

        :param path: the relative path of the directory.
        :return: the `File` representing the object located at the given relative path.
        :raises TypeError: if the path is not a string or ``Path``, or is an absolute path.
        :raises FileNotFoundError: if no object exists for the given path.
        :raises NotADirectoryError: if the object at the given path is not a directory.
        """
        file_object = self.get_object(path)

        if file_object.file_type != FileType.DIRECTORY:
            raise NotADirectoryError(f'object with path `{path}` is not a directory.')

        return file_object

    def get_file(self, path: FilePath) -> File:
        """Return the file object at the given path.

        :param path: the relative path of the file object.
        :return: the `File` representing the object located at the given relative path.
        :raises TypeError: if the path is not a string or ``Path``, or is an absolute path.
        :raises FileNotFoundError: if no object exists for the given path.
        :raises IsADirectoryError: if the object at the given path is not a directory.
        """
        if path is None:
            raise TypeError('path cannot be `None`.')

        path = self._pre_process_path(path)

        file_object = self.get_object(path)

        if file_object.file_type != FileType.FILE:
            raise IsADirectoryError(f'object with path `{path}` is not a file.')

        return file_object

    def list_objects(self, path: FilePath = None) -> List[File]:
        """Return a list of the objects contained in this repository sorted by name, optionally in given sub directory.

        :param path: the relative path of the directory.
        :return: a list of `File` named tuples representing the objects present in directory with the given path.
        :raises TypeError: if the path is not a string or ``Path``, or is an absolute path.
        :raises FileNotFoundError: if no object exists for the given path.
        :raises NotADirectoryError: if the object at the given path is not a directory.
        """
        directory = self.get_directory(path)
        return sorted(directory.objects.values(), key=lambda obj: obj.name)

    def list_object_names(self, path: FilePath = None) -> List[str]:
        """Return a sorted list of the object names contained in this repository, optionally in the given sub directory.

        :param path: the relative path of the directory.
        :return: a list of `File` named tuples representing the objects present in directory with the given path.
        :raises TypeError: if the path is not a string or ``Path``, or is an absolute path.
        :raises FileNotFoundError: if no object exists for the given path.
        :raises NotADirectoryError: if the object at the given path is not a directory.
        """
        return [entry.name for entry in self.list_objects(path)]

    def put_object_from_filelike(self, handle: BinaryIO, path: FilePath) -> None:
        """Store the byte contents of a file in the repository.

        :param handle: filelike object with the byte content to be stored.
        :param path: the relative path where to store the object in the repository.
        :raises TypeError: if the path is not a string or ``Path``, or is an absolute path.
        """
        path = self._pre_process_path(path)
        key = self.backend.put_object_from_filelike(handle)
        self._insert_file(path, key)

    def put_object_from_file(self, filepath: FilePath, path: FilePath) -> None:
        """Store a new object under `path` with contents of the file located at `filepath` on the local file system.

        :param filepath: absolute path of file whose contents to copy to the repository
        :param path: the relative path where to store the object in the repository.
        :raises TypeError: if the path is not a string and relative path, or the handle is not a byte stream.
        """
        with open(filepath, 'rb') as handle:
            self.put_object_from_filelike(handle, path)

    def put_object_from_tree(self, filepath: FilePath, path: FilePath = None) -> None:
        """Store the entire contents of `filepath` on the local file system in the repository with under given `path`.

        :param filepath: absolute path of the directory whose contents to copy to the repository.
        :param path: the relative path where to store the objects in the repository.
        :raises TypeError: if the filepath is not a string or ``Path``, or is a relative path.
        :raises TypeError: if the path is not a string or ``Path``, or is an absolute path.
        """
        import os

        path = self._pre_process_path(path)

        if isinstance(filepath, str):
            filepath = pathlib.PurePosixPath(filepath)

        if not isinstance(filepath, pathlib.PurePosixPath):
            raise TypeError(f'filepath `{filepath}` is not of type `str` nor `pathlib.PurePosixPath`.')

        if not filepath.is_absolute():
            raise TypeError(f'filepath `{filepath}` is not an absolute path.')

        # Explicitly create the base directory if specified by `path`, just in case `filepath` contains no file objects.
        if path.parts:
            self.create_directory(path)

        for root_str, dirnames, filenames in os.walk(filepath):

            root = pathlib.PurePosixPath(root_str)

            for dirname in dirnames:
                self.create_directory(path / root.relative_to(filepath) / dirname)

            for filename in filenames:
                self.put_object_from_file(root / filename, path / root.relative_to(filepath) / filename)

    def is_empty(self) -> bool:
        """Return whether the repository is empty.

        :return: True if the repository contains no file objects.
        """
        return not self._directory.objects

    def has_object(self, path: FilePath) -> bool:
        """Return whether the repository has an object with the given path.

        :param path: the relative path of the object within the repository.
        :return: True if the object exists, False otherwise.
        :raises TypeError: if the path is not a string or ``Path``, or is an absolute path.
        """
        try:
            self.get_object(path)
        except FileNotFoundError:
            return False
        else:
            return True

    @contextlib.contextmanager
    def open(self, path: FilePath) -> Iterator[BinaryIO]:
        """Open a file handle to an object stored under the given path.

        .. note:: this should only be used to open a handle to read an existing file. To write a new file use the method
            ``put_object_from_filelike`` instead.

        :param path: the relative path of the object within the repository.
        :return: yield a byte stream object.
        :raises TypeError: if the path is not a string or ``Path``, or is an absolute path.
        :raises FileNotFoundError: if the file does not exist.
        :raises IsADirectoryError: if the object is a directory and not a file.
        :raises OSError: if the file could not be opened.
        """
        key = self.get_file(path).key
        assert key is not None, 'Expected FileType.File to have a key'
        with self.backend.open(key) as handle:
            yield handle

    def get_object_content(self, path: FilePath) -> bytes:
        """Return the content of a object identified by path.

        :param path: the relative path of the object within the repository.
        :raises TypeError: if the path is not a string or ``Path``, or is an absolute path.
        :raises FileNotFoundError: if the file does not exist.
        :raises IsADirectoryError: if the object is a directory and not a file.
        :raises OSError: if the file could not be opened.
        """
        key = self.get_file(path).key
        assert key is not None, 'Expected FileType.File to have a key'
        return self.backend.get_object_content(key)

    def delete_object(self, path: FilePath, hard_delete: bool = False) -> None:
        """Soft delete the object from the repository.

        .. note:: can only delete file objects, but not directories.

        :param path: the relative path of the object within the repository.
        :param hard_delete: when true, not only remove the file from the internal mapping but also call through to the
            ``delete_object`` method of the actual repository backend.
        :raises TypeError: if the path is not a string or ``Path``, or is an absolute path.
        :raises FileNotFoundError: if the file does not exist.
        :raises IsADirectoryError: if the object is a directory and not a file.
        :raises OSError: if the file could not be deleted.
        """
        path = self._pre_process_path(path)
        file_object = self.get_object(path)

        if file_object.file_type == FileType.DIRECTORY:
            raise IsADirectoryError(f'object with path `{path}` is a directory.')

        if hard_delete:
            assert file_object.key is not None, 'Expected FileType.File to have a key'
            self.backend.delete_object(file_object.key)

        directory = self.get_directory(path.parent)
        directory.objects.pop(path.name)

    def erase(self) -> None:
        """Delete all objects from the repository.

        .. important: this intentionally does not call through to any ``erase`` method of the backend, because unlike
            this class, the backend does not just store the objects of a single node, but potentially of a lot of other
            nodes. Therefore, we manually delete all file objects and then simply reset the internal file hierarchy.

        """
        for file_key in self.get_file_keys():
            self.backend.delete_object(file_key)
        self.reset()

    def clone(self, source: 'Repository') -> None:
        """Clone the contents of another repository instance."""
        if not isinstance(source, Repository):
            raise TypeError('source is not an instance of `Repository`.')

        for root, dirnames, filenames in source.walk():
            for dirname in dirnames:
                self.create_directory(root / dirname)
            for filename in filenames:
                with source.open(root / filename) as handle:
                    self.put_object_from_filelike(handle, root / filename)

    def walk(self, path: FilePath = None) -> Iterable[Tuple[pathlib.PurePosixPath, List[str], List[str]]]:
        """Walk over the directories and files contained within this repository.

        .. note:: the order of the dirname and filename lists that are returned is not necessarily sorted. This is in
            line with the ``os.walk`` implementation where the order depends on the underlying file system used.

        :param path: the relative path of the directory within the repository whose contents to walk.
        :return: tuples of root, dirnames and filenames just like ``os.walk``, with the exception that the root path is
            always relative with respect to the repository root, instead of an absolute path and it is an instance of
            ``pathlib.PurePosixPath`` instead of a normal string
        """
        path = self._pre_process_path(path)

        directory = self.get_directory(path)
        dirnames = [obj.name for obj in directory.objects.values() if obj.file_type == FileType.DIRECTORY]
        filenames = [obj.name for obj in directory.objects.values() if obj.file_type == FileType.FILE]

        if dirnames:
            for dirname in dirnames:
                yield from self.walk(path / dirname)

        yield path, dirnames, filenames

    def copy_tree(self, target: Union[str, pathlib.Path], path: FilePath = None) -> None:
        """Copy the contents of the entire node repository to another location on the local file system.

        :param target: absolute path of the directory where to copy the contents to.
        :param path: optional relative path whose contents to copy.
        :raises TypeError: if ``target`` is of incorrect type or not absolute.
        :raises NotADirectoryError: if ``path`` does not reference a directory.
        """
        path = self._pre_process_path(path)
        file_object = self.get_object(path)

        if file_object.file_type != FileType.DIRECTORY:
            raise NotADirectoryError(f'object with path `{path}` is not a directory.')

        if isinstance(target, str):
            target = pathlib.Path(target)

        if not isinstance(target, pathlib.Path):
            raise TypeError(f'path `{path}` is not of type `str` nor `pathlib.Path`.')

        if not target.is_absolute():
            raise TypeError(f'provided target `{target}` is not an absolute path.')

        for root, dirnames, filenames in self.walk(path):
            for dirname in dirnames:
                dirpath = target / root / dirname
                dirpath.mkdir(parents=True, exist_ok=True)

            for filename in filenames:
                dirpath = target / root
                filepath = dirpath / filename

                dirpath.mkdir(parents=True, exist_ok=True)

                with self.open(root / filename) as handle:
                    filepath.write_bytes(handle.read())

    # these methods are not actually used in aiida-core, but are here for completeness

    def initialise(self, **kwargs: Any) -> None:
        """Initialise the repository if it hasn't already been initialised.

        :param kwargs: keyword argument that will be passed to the ``initialise`` call of the backend.
        """
        self.backend.initialise(**kwargs)

    def delete(self) -> None:
        """Delete the repository.

        .. important:: This will not just delete the contents of the repository but also the repository itself and all
            of its assets. For example, if the repository is stored inside a folder on disk, the folder may be deleted.
        """
        self.backend.erase()
        self.reset()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Updating link types - This is a copy of the Django migration script

Revision ID: a6048f0ffca8
Revises:
Create Date: 2017-10-17 10:51:23.327195

"""
from alembic import op
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = 'a6048f0ffca8'
down_revision = '70c7d732f1b2'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    conn = op.get_bind()

    # I am first migrating the wrongly declared returnlinks out of
    # the InlineCalculations.
    # This bug is reported #628 https://github.com/aiidateam/aiida-core/issues/628
    # There is an explicit check in the code of the inline calculation
    # ensuring that the calculation returns UNSTORED nodes.
    # Therefore, no cycle can be created with that migration!
    #
    # this command:
    # 1) selects all links that
    #   - joins an InlineCalculation (or subclass) as input
    #   - joins a Data (or subclass) as output
    #   - is marked as a returnlink.
    # 2) set for these links the type to 'createlink'
    stmt1 = text(
        """
        UPDATE db_dblink set type='createlink' WHERE db_dblink.id IN (
            SELECT db_dblink_1.id
            FROM db_dbnode AS db_dbnode_1
                JOIN db_dblink AS db_dblink_1 ON db_dblink_1.input_id = db_dbnode_1.id
                JOIN db_dbnode AS db_dbnode_2 ON db_dblink_1.output_id = db_dbnode_2.id
            WHERE db_dbnode_1.type LIKE 'calculation.inline.%'
                AND db_dbnode_2.type LIKE 'data.%'
                AND db_dblink_1.type = 'returnlink'
        )
    """
    )
    conn.execute(stmt1)
    # Now I am updating the link-types that are null because of either an export and subsequent import
    # https://github.com/aiidateam/aiida-core/issues/685
    # or because the link types don't exist because the links were added before the introduction of link types.
    # This is reported here: https://github.com/aiidateam/aiida-core/issues/687
    #
    # The following sql statement:
    # 1) selects all links that
    #   - joins Data (or subclass) or Code as input
    #   - joins Calculation (or subclass) as output. This includes WorkCalculation, InlineCalcuation, JobCalculations...
    #   - has no type (null)
    # 2) set for these links the type to 'inputlink'
    stmt2 = text(
        """
         UPDATE db_dblink set type='inputlink' where id in (
            SELECT db_dblink_1.id
            FROM db_dbnode AS db_dbnode_1
                JOIN db_dblink AS db_dblink_1 ON db_dblink_1.input_id = db_dbnode_1.id
                JOIN db_dbnode AS db_dbnode_2 ON db_dblink_1.output_id = db_dbnode_2.id
            WHERE ( db_dbnode_1.type LIKE 'data.%' or db_dbnode_1.type = 'code.Code.' )
                AND db_dbnode_2.type LIKE 'calculation.%'
                AND ( db_dblink_1.type = null OR db_dblink_1.type = '')
        );
    """
    )
    conn.execute(stmt2)
    #
    # The following sql statement:
    # 1) selects all links that
    #   - join JobCalculation (or subclass) or InlineCalculation as input
    #   - joins Data (or subclass) as output.
    #   - has no type (null)
    # 2) set for these links the type to 'createlink'
    stmt3 = text(
        """
         UPDATE db_dblink set type='createlink' where id in (
            SELECT db_dblink_1.id
            FROM db_dbnode AS db_dbnode_1
                JOIN db_dblink AS db_dblink_1 ON db_dblink_1.input_id = db_dbnode_1.id
                JOIN db_dbnode AS db_dbnode_2 ON db_dblink_1.output_id = db_dbnode_2.id
            WHERE db_dbnode_2.type LIKE 'data.%'
                AND (
                    db_dbnode_1.type LIKE 'calculation.job.%'
                    OR
                    db_dbnode_1.type = 'calculation.inline.InlineCalculation.'
                )
                AND ( db_dblink_1.type = null OR db_dblink_1.type = '')
        )
    """
    )
    conn.execute(stmt3)
    # The following sql statement:
    # 1) selects all links that
    #   - join WorkCalculation as input. No subclassing was introduced so far, so only one type string is checked for.
    #   - join Data (or subclass) as output.
    #   - has no type (null)
    # 2) set for these links the type to 'returnlink'
    stmt4 = text(
        """
         UPDATE db_dblink set type='returnlink' where id in (
            SELECT db_dblink_1.id
            FROM db_dbnode AS db_dbnode_1
                JOIN db_dblink AS db_dblink_1 ON db_dblink_1.input_id = db_dbnode_1.id
                JOIN db_dbnode AS db_dbnode_2 ON db_dblink_1.output_id = db_dbnode_2.id
            WHERE db_dbnode_2.type LIKE 'data.%'
                AND db_dbnode_1.type = 'calculation.work.WorkCalculation.'
                AND ( db_dblink_1.type = null OR db_dblink_1.type = '')
        )
    """
    )
    conn.execute(stmt4)
    # Now I update links that are CALLS:
    # The following sql statement:
    # 1) selects all links that
    #   - join WorkCalculation as input. No subclassing was introduced so far, so only one type string is checked for.
    #   - join Calculation (or subclass) as output. Includes JobCalculation and WorkCalculations and all subclasses.
    #   - has no type (null)
    # 2) set for these links the type to 'calllink'
    stmt5 = text(
        """
         UPDATE db_dblink set type='calllink' where id in (
            SELECT db_dblink_1.id
            FROM db_dbnode AS db_dbnode_1
                JOIN db_dblink AS db_dblink_1 ON db_dblink_1.input_id = db_dbnode_1.id
                JOIN db_dbnode AS db_dbnode_2 ON db_dblink_1.output_id = db_dbnode_2.id
            WHERE db_dbnode_1.type = 'calculation.work.WorkCalculation.'
                AND db_dbnode_2.type LIKE 'calculation.%'
                AND ( db_dblink_1.type = null  OR db_dblink_1.type = '')
        )
    """
    )
    conn.execute(stmt5)


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of a6048f0ffca8.')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Migration to add the `repository_metadata` JSONB column.

Revision ID: 7536a82b2cc4
Revises: 0edcdd5a30f0
Create Date: 2020-07-09 11:32:39.924151

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '7536a82b2cc4'
down_revision = '0edcdd5a30f0'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    op.add_column(
        'db_dbnode',
        sa.Column('repository_metadata', postgresql.JSONB(astext_type=sa.Text()), nullable=False, server_default='{}')
    )
    op.alter_column('db_dbnode', 'repository_metadata', server_default=None)


def downgrade():
    """Migrations for the downgrade."""
    op.drop_column('db_dbnode', 'repository_metadata')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""SQLA user"""
from aiida.orm.implementation.users import BackendUser, BackendUserCollection
from aiida.storage.psql_dos.models.user import DbUser

from . import entities, utils


class SqlaUser(entities.SqlaModelEntity[DbUser], BackendUser):
    """SQLA user"""

    MODEL_CLASS = DbUser

    def __init__(self, backend, email, first_name, last_name, institution):
        # pylint: disable=too-many-arguments
        super().__init__(backend)
        self._model = utils.ModelWrapper(
            self.MODEL_CLASS(email=email, first_name=first_name, last_name=last_name, institution=institution), backend
        )

    @property
    def email(self):
        return self.model.email

    @email.setter
    def email(self, email):
        self.model.email = email

    @property
    def first_name(self):
        return self.model.first_name

    @first_name.setter
    def first_name(self, first_name):
        self.model.first_name = first_name

    @property
    def last_name(self):
        return self.model.last_name

    @last_name.setter
    def last_name(self, last_name):
        self.model.last_name = last_name

    @property
    def institution(self):
        return self.model.institution

    @institution.setter
    def institution(self, institution):
        self.model.institution = institution


class SqlaUserCollection(BackendUserCollection):
    """Collection of SQLA Users"""

    ENTITY_CLASS = SqlaUser

    def create(self, email, first_name='', last_name='', institution=''):  # pylint: disable=arguments-differ
        """ Create a user with the provided email address"""
        return self.ENTITY_CLASS(self.backend, email, first_name, last_name, institution)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Shared function for django_0024 and sqlalchemy ea2f50e7f615"""
import sys
from tempfile import NamedTemporaryFile

import click
import sqlalchemy as sa

from aiida.cmdline.utils import echo

from .utils import dumps_json


def get_legacy_workflow_log_number(connection):
    """ Get the number of the log records that correspond to legacy workflows """
    return connection.execute(
        sa.text(
            """
        SELECT COUNT(*) FROM db_dblog
        WHERE
            (db_dblog.objname LIKE 'aiida.workflows.user.%')
        """
        )
    ).fetchall()[0][0]


def get_unknown_entity_log_number(connection):
    """ Get the number of the log records that correspond to unknown entities """
    return connection.execute(
        sa.text(
            """
        SELECT COUNT(*) FROM db_dblog
        WHERE
            (db_dblog.objname NOT LIKE 'node.%') AND
            (db_dblog.objname NOT LIKE 'aiida.workflows.user.%')
        """
        )
    ).fetchall()[0][0]


def get_logs_with_no_nodes_number(connection):
    """ Get the number of the log records that correspond to nodes that were deleted """
    return connection.execute(
        sa.text(
            """
        SELECT COUNT(*) FROM db_dblog
        WHERE
            (db_dblog.objname LIKE 'node.%') AND NOT EXISTS
            (SELECT 1 FROM db_dbnode WHERE db_dbnode.id = db_dblog.objpk LIMIT 1)
        """
        )
    ).fetchall()[0][0]


def get_serialized_legacy_workflow_logs(connection):
    """ Get the serialized log records that correspond to legacy workflows """
    query = connection.execute(
        sa.text(
            """
        SELECT db_dblog.id, db_dblog.time, db_dblog.loggername, db_dblog.levelname, db_dblog.objpk, db_dblog.objname,
        db_dblog.message, db_dblog.metadata FROM db_dblog
        WHERE
            (db_dblog.objname LIKE 'aiida.workflows.user.%')
        """
        )
    )
    res = []
    for row in query:
        res.append(row._asdict())
    return dumps_json(res)


def get_serialized_unknown_entity_logs(connection):
    """ Get the serialized log records that correspond to unknown entities """
    query = connection.execute(
        sa.text(
            """
        SELECT db_dblog.id, db_dblog.time, db_dblog.loggername, db_dblog.levelname, db_dblog.objpk, db_dblog.objname,
        db_dblog.message, db_dblog.metadata FROM db_dblog
        WHERE
            (db_dblog.objname NOT LIKE 'node.%') AND
            (db_dblog.objname NOT LIKE 'aiida.workflows.user.%')
        """
        )
    )
    res = []
    for row in query:
        res.append(row._asdict())
    return dumps_json(res)


def get_serialized_logs_with_no_nodes(connection):
    """ Get the serialized log records that correspond to nodes that were deleted """
    query = connection.execute(
        sa.text(
            """
        SELECT db_dblog.id, db_dblog.time, db_dblog.loggername, db_dblog.levelname, db_dblog.objpk, db_dblog.objname,
        db_dblog.message, db_dblog.metadata FROM db_dblog
        WHERE
            (db_dblog.objname LIKE 'node.%') AND NOT EXISTS
            (SELECT 1 FROM db_dbnode WHERE db_dbnode.id = db_dblog.objpk LIMIT 1)
        """
        )
    )
    res = []
    for row in query:
        res.append(row._asdict())
    return dumps_json(res)


def export_and_clean_workflow_logs(connection, profile):
    """Export the logs records that correspond to legacy workflows and to unknown entities
    (place them to files and remove them from the DbLog table).
    """
    lwf_no_number = get_legacy_workflow_log_number(connection)
    other_number = get_unknown_entity_log_number(connection)
    log_no_node_number = get_logs_with_no_nodes_number(connection)

    # If there are no legacy workflow log records or log records of an unknown entity
    if lwf_no_number == 0 and other_number == 0 and log_no_node_number == 0:
        return

    if not profile.is_test_profile:
        echo.echo_warning(
            'We found {} log records that correspond to legacy workflows and {} log records to correspond '
            'to an unknown entity.'.format(lwf_no_number, other_number)
        )
        echo.echo_warning(
            'These records will be removed from the database and exported to JSON files (to the current directory).'
        )
        proceed = click.confirm('Would you like to proceed?', default=True)
        if not proceed:
            sys.exit(1)

    delete_on_close = profile.is_test_profile

    # Exporting the legacy workflow log records
    if lwf_no_number != 0:

        # Get the records and write them to file
        with NamedTemporaryFile(
            prefix='legagy_wf_logs-', suffix='.log', dir='.', delete=delete_on_close, mode='w+'
        ) as handle:
            # Export the log records
            filename = handle.name
            handle.write(get_serialized_legacy_workflow_logs(connection))

            # If delete_on_close is False, we are running for the user and add additional message of file location
            if not delete_on_close:
                echo.echo(f'Exported legacy workflow logs to {filename}')

        # Now delete the records
        connection.execute(
            sa.text(
                """
            DELETE FROM db_dblog
            WHERE
                (db_dblog.objname LIKE 'aiida.workflows.user.%')
            """
            )
        )

    # Exporting unknown log records
    if other_number != 0:
        # Get the records and write them to file
        with NamedTemporaryFile(
            prefix='unknown_entity_logs-', suffix='.log', dir='.', delete=delete_on_close, mode='w+'
        ) as handle:
            # Export the log records
            filename = handle.name
            handle.write(get_serialized_unknown_entity_logs(connection))

            # If delete_on_close is False, we are running for the user and add additional message of file location
            if not delete_on_close:
                echo.echo(f'Exported unexpected entity logs to {filename}')

        # Now delete the records
        connection.execute(
            sa.text(
                """
            DELETE FROM db_dblog WHERE
                (db_dblog.objname NOT LIKE 'node.%') AND
                (db_dblog.objname NOT LIKE 'aiida.workflows.user.%')
            """
            )
        )

    # Exporting log records that don't correspond to nodes
    if log_no_node_number != 0:
        # Get the records and write them to file
        with NamedTemporaryFile(
            prefix='no_node_entity_logs-', suffix='.log', dir='.', delete=delete_on_close, mode='w+'
        ) as handle:
            # Export the log records
            filename = handle.name
            handle.write(get_serialized_logs_with_no_nodes(connection))

            # If delete_on_close is False, we are running for the user and add additional message of file location
            if not delete_on_close:
                echo.echo(f'Exported entity logs that do not correspond to nodes to {filename}')

        # Now delete the records
        connection.execute(
            sa.text(
                """
            DELETE FROM db_dblog WHERE
            (db_dblog.objname LIKE 'node.%') AND NOT EXISTS
            (SELECT 1 FROM db_dbnode WHERE db_dbnode.id = db_dblog.objpk LIMIT 1)
            """
            )
        )


def set_new_uuid(connection):
    """ Set new and distinct UUIDs to all the logs """
    from aiida.common.utils import get_new_uuid

    # Exit if there are no rows - e.g. initial setup
    id_query = connection.execute(sa.text('SELECT db_dblog.id FROM db_dblog'))
    if id_query.rowcount == 0:
        return

    id_res = id_query.fetchall()
    ids = []
    for (curr_id,) in id_res:
        ids.append(curr_id)
    uuids = set()
    while len(uuids) < len(ids):
        uuids.add(get_new_uuid())

    # Create the key/value pairs
    key_values = ','.join(f"({curr_id}, '{curr_uuid}')" for curr_id, curr_uuid in zip(ids, uuids))

    update_stm = f"""
        UPDATE db_dblog as t SET
            uuid = uuid(c.uuid)
        from (values {key_values}) as c(id, uuid) where c.id = t.id"""
    connection.execute(sa.text(update_stm))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Definition of the ``SqliteTempBackend`` backend."""
from __future__ import annotations

from contextlib import contextmanager
from functools import cached_property
import os
from pathlib import Path
from typing import Any, Iterator, Sequence

from sqlalchemy.orm import Session

from aiida.common.exceptions import ClosedStorage
from aiida.manage import Profile
from aiida.manage.configuration.settings import AIIDA_CONFIG_FOLDER
from aiida.orm.entities import EntityTypes
from aiida.orm.implementation import BackendEntity, StorageBackend
from aiida.repository.backend.sandbox import SandboxRepositoryBackend
from aiida.storage.sqlite_zip import models, orm
from aiida.storage.sqlite_zip.migrator import get_schema_version_head
from aiida.storage.sqlite_zip.utils import create_sqla_engine

__all__ = ('SqliteTempBackend',)


class SqliteTempBackend(StorageBackend):  # pylint: disable=too-many-public-methods
    """A temporary backend, using an in-memory sqlite database.

    This backend is intended for testing and demonstration purposes.
    Whenever it is instantiated, it creates a fresh storage backend,
    and destroys it when it is garbage collected.
    """
    _read_only = False

    @staticmethod
    def create_profile(
        name: str = 'temp',
        default_user_email='user@email.com',
        sandbox_path: str | Path | None = None,
        options: dict | None = None,
        debug: bool = False
    ) -> Profile:
        """Create a new profile instance for this backend, from the path to the zip file."""
        sandbox_path = sandbox_path or os.path.join(AIIDA_CONFIG_FOLDER, 'repository', name)
        return Profile(
            name,
            {
                'default_user_email': default_user_email,
                'storage': {
                    'backend': 'sqlite_temp',
                    'config': {
                        'debug': debug,
                        # Note this is currently required, see https://github.com/aiidateam/aiida-core/issues/5451
                        'repository_uri': f'file://{os.path.abspath(sandbox_path)}',
                    }
                },
                'process_control': {
                    'backend': 'null',
                    'config': {}
                },
                'options': options or {},
            }
        )

    @classmethod
    def version_head(cls) -> str:
        return get_schema_version_head()

    @classmethod
    def version_profile(cls, profile: Profile) -> str | None:  # pylint: disable=unused-argument
        return get_schema_version_head()

    @classmethod
    def migrate(cls, profile: Profile):
        pass

    def __init__(self, profile: Profile):
        super().__init__(profile)
        self._session: Session | None = None
        self._repo: SandboxRepositoryBackend | None = None
        self._globals: dict[str, tuple[Any, str | None]] = {}
        self._closed = False
        self.get_session()  # load the database on initialization

    def __str__(self) -> str:
        state = 'closed' if self.is_closed else 'open'
        return f'SqliteTemp storage [{state}], sandbox: {self.profile.repository_path}'

    @property
    def is_closed(self) -> bool:
        return self._closed

    def close(self):
        if self._session:
            self._session.close()
        if self._repo:
            self._repo.erase()
        self._session = None
        self._repo = None
        self._globals = {}
        self._closed = True

    def get_global_variable(self, key: str):
        return self._globals[key][0]

    def set_global_variable(self, key: str, value, description: str | None = None, overwrite=True) -> None:  # pylint: disable=unused-argument
        if not overwrite and key in self._globals:
            raise ValueError(f'global variable {key} already exists')
        self._globals[key] = (value, description)

    def get_session(self) -> Session:
        """Return an SQLAlchemy session."""
        if self._closed:
            raise ClosedStorage(str(self))
        if self._session is None:
            engine = create_sqla_engine(':memory:', echo=self.profile.storage_config.get('debug', False))
            models.SqliteBase.metadata.create_all(engine)
            self._session = Session(engine, future=True)
            self._session.add(models.DbUser(email=self.profile.default_user_email or 'user@email.com'))
            self._session.commit()
        return self._session

    def get_repository(self) -> SandboxRepositoryBackend:
        if self._closed:
            raise ClosedStorage(str(self))
        if self._repo is None:
            # to-do this does not seem to be removing the folder on garbage collection?
            self._repo = SandboxRepositoryBackend(sandbox_in_repo=True)
        return self._repo

    @property
    def in_transaction(self) -> bool:
        return self.get_session().in_nested_transaction()

    @contextmanager
    def transaction(self) -> Iterator[Session]:
        """Open a transaction to be used as a context manager.

        If there is an exception within the context then the changes will be rolled back and the state will be as before
        entering. Transactions can be nested.
        """
        session = self.get_session()
        if session.in_transaction():
            with session.begin_nested():
                yield session
            session.commit()
        else:
            with session.begin():
                with session.begin_nested():
                    yield session

    def _clear(self, recreate_user: bool = True) -> None:
        raise NotImplementedError

    def maintain(self, dry_run: bool = False, live: bool = True, **kwargs) -> None:
        pass

    def query(self) -> orm.SqliteQueryBuilder:
        return orm.SqliteQueryBuilder(self)

    def get_backend_entity(self, model) -> BackendEntity:
        """Return the backend entity that corresponds to the given Model instance."""
        return orm.get_backend_entity(model, self)

    @cached_property
    def authinfos(self):
        return orm.SqliteAuthInfoCollection(self)

    @cached_property
    def comments(self):
        return orm.SqliteCommentCollection(self)

    @cached_property
    def computers(self):
        return orm.SqliteComputerCollection(self)

    @cached_property
    def groups(self):
        return orm.SqliteGroupCollection(self)

    @cached_property
    def logs(self):
        return orm.SqliteLogCollection(self)

    @cached_property
    def nodes(self):
        return orm.SqliteNodeCollection(self)

    @cached_property
    def users(self):
        return orm.SqliteUserCollection(self)

    def get_info(self, detailed: bool = False) -> dict:
        results = super().get_info(detailed=detailed)
        # results['repository'] = self.get_repository().get_info(detailed)
        return results

    def bulk_insert(self, entity_type: EntityTypes, rows: list[dict], allow_defaults: bool = False) -> list[int]:
        raise NotImplementedError

    def bulk_update(self, entity_type: EntityTypes, rows: list[dict]) -> None:
        raise NotImplementedError

    def delete_nodes_and_connections(self, pks_to_delete: Sequence[int]):
        raise NotImplementedError
